@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -18,6 +18,7 @@
  */
 
 #include <osgEarth/ImageUtils>
+#include <osgEarth/ThreadingUtils>
 #include <osg/Notify>
 #include <osg/Texture>
 #include <osg/ImageSequence>
@@ -28,6 +29,16 @@
 
 #define LC "[ImageUtils] "
 
+
+#if defined(OSG_GLES1_AVAILABLE) || defined(OSG_GLES2_AVAILABLE)
+#    define GL_RGB8_INTERNAL  GL_RGB8_OES
+#    define GL_RGB8A_INTERNAL GL_RGBA8_OES
+#else
+#    define GL_RGB8_INTERNAL  GL_RGB8
+#    define GL_RGB8A_INTERNAL GL_RGBA8
+#endif
+
+
 using namespace osgEarth;
 
 osg::Image*
@@ -39,31 +50,10 @@ ImageUtils::cloneImage( const osg::Image* input )
     // Calling clone->dirty() might work, but we are not sure.
 
     if ( !input ) return 0L;
-
-    if ( input->className() == "Image" )
-    {
-#if 0
-        osg::Image* clone = new osg::Image( *input );
-        clone->dirty();
-        return clone;
-#else
-        osg::Image* clone = new osg::Image();
-        clone->allocateImage( input->s(), input->t(), input->r(), input->getPixelFormat(), input->getDataType(), input->getPacking() );
-        clone->setInternalTextureFormat( input->getInternalTextureFormat() );
-        if ( input->isMipmap() )
-            clone->setMipmapLevels( input->getMipmapLevels() );
-        memcpy( clone->data(), input->data(), input->getTotalSizeInBytesIncludingMipmaps() );
-        return clone;
-#endif
-    }
-
-    else
-    {
-        // handles Image subclasses.
-        osg::Image* clone = osg::clone( input, osg::CopyOp::DEEP_COPY_ALL );
-        clone->dirty();
-        return clone;
-    }
+    
+    osg::Image* clone = osg::clone( input, osg::CopyOp::DEEP_COPY_ALL );
+    clone->dirty();
+    return clone;
 }
 
 void
@@ -75,9 +65,9 @@ ImageUtils::normalizeImage( osg::Image* image )
     if ( image->getDataType() == GL_UNSIGNED_BYTE )
     {
         if ( image->getPixelFormat() == GL_RGB )
-            image->setInternalTextureFormat( GL_RGB8 );
+            image->setInternalTextureFormat( GL_RGB8_INTERNAL );
         else if ( image->getPixelFormat() == GL_RGBA )
-            image->setInternalTextureFormat( GL_RGBA8 );
+            image->setInternalTextureFormat( GL_RGB8A_INTERNAL );
     }
 }
 
@@ -123,6 +113,56 @@ ImageUtils::copyAsSubImage(const osg::Image* src, osg::Image* dst, int dst_start
     return true;
 }  
 
+osg::Image*
+ImageUtils::createBumpMap(const osg::Image* input)
+{
+    osg::Image* output = osg::clone(input, osg::CopyOp::DEEP_COPY_ALL);
+
+    static const float kernel[] = {
+        -1.0, -1.0, 0.0,
+        -1.0,  0.0, 1.0,
+         0.0,  1.0, 1.0 
+    };
+
+    PixelReader read(input);
+    PixelWriter write(output);
+
+    osg::Vec4f mid(0.5f,0.5f,0.5f,0.5f);
+
+    for( int t=0; t<input->t(); ++t )
+    {
+        for( int s=0; s<input->s(); ++s )
+        {
+            if ( t == 0 || t == input->t()-1 || s == 0 || s == input->s()-1 )
+            {
+                write( mid, s, t );
+            }
+            else
+            {
+                osg::Vec4f sum;
+
+                // run the emboss kernel:
+                for( int tt=0; tt<=2; ++tt )
+                    for( int ss=0; ss<=2; ++ss )
+                        sum += read(s+ss-1,t+tt-1) * kernel[tt*3+ss];
+                sum /= 9.0f;
+
+                // bias for bumpmapping:
+                sum += osg::Vec4f(0.5f,0.5f,0.5f,0.5f);
+
+                // convert to greyscale:
+                sum.r() *= 0.2989f;
+                sum.g() *= 0.5870f;
+                sum.b() *= 0.1140f;
+
+                sum.a() = read(s,t).a();
+                write( sum, s, t );
+            }
+        }
+    }
+    return output;
+}
+
 bool
 ImageUtils::resizeImage(const osg::Image* input, 
                         unsigned int out_s, unsigned int out_t, 
@@ -160,7 +200,7 @@ ImageUtils::resizeImage(const osg::Image* input,
         {
             // for unsupported write formats, convert to RGBA8 automatically.
             output->allocateImage( out_s, out_t, 1, GL_RGBA, GL_UNSIGNED_BYTE );
-            output->setInternalTextureFormat( GL_RGBA8 );
+            output->setInternalTextureFormat( GL_RGB8A_INTERNAL );
         }
     }
     else
@@ -277,7 +317,7 @@ namespace
                 dest.r()*(1.0f-sa) + src.r()*sa,
                 dest.g()*(1.0f-sa) + src.g()*sa,
                 dest.b()*(1.0f-sa) + src.b()*sa,
-                dest.a() );
+                osg::maximum(sa, da) );             
             return true;
         }
     };
@@ -292,9 +332,9 @@ ImageUtils::mix(osg::Image* dest, const osg::Image* src, float a)
     PixelVisitor<MixImage> mixer;
     mixer._a = osg::clampBetween( a, 0.0f, 1.0f );
     mixer._srcHasAlpha = src->getPixelSizeInBits() == 32;
-    mixer._destHasAlpha = src->getPixelSizeInBits() == 32;
+    mixer._destHasAlpha = src->getPixelSizeInBits() == 32;    
 
-    mixer.accept( src, dest );
+    mixer.accept( src, dest );  
 
     return true;
 }
@@ -304,6 +344,9 @@ ImageUtils::cropImage(const osg::Image* image,
                       double src_minx, double src_miny, double src_maxx, double src_maxy,
                       double &dst_minx, double &dst_miny, double &dst_maxx, double &dst_maxy)
 {
+    if ( image == 0L )
+        return 0L;
+
     //Compute the desired cropping rectangle
     int windowX        = osg::clampBetween( (int)floor( (dst_minx - src_minx) / (src_maxx - src_minx) * (double)image->s()), 0, image->s()-1);
     int windowY        = osg::clampBetween( (int)floor( (dst_miny - src_miny) / (src_maxy - src_miny) * (double)image->t()), 0, image->t()-1);
@@ -393,17 +436,89 @@ ImageUtils::sharpenImage( const osg::Image* input )
     return output;
 }
 
+namespace
+{
+    static Threading::Mutex         s_emptyImageMutex;
+    static osg::ref_ptr<osg::Image> s_emptyImage;
+}
 
 osg::Image*
 ImageUtils::createEmptyImage()
+{
+    if (!s_emptyImage.valid())
+    {
+        Threading::ScopedMutexLock exclusive( s_emptyImageMutex );
+        if (!s_emptyImage.valid())
+        {
+            s_emptyImage = new osg::Image;
+            s_emptyImage->allocateImage(1,1,1, GL_RGBA, GL_UNSIGNED_BYTE);
+            s_emptyImage->setInternalTextureFormat( GL_RGB8A_INTERNAL );
+            unsigned char *data = s_emptyImage->data(0,0);
+            memset(data, 0, 4);
+        }     
+    }
+    return s_emptyImage.get();
+}
+
+bool
+ImageUtils::isEmptyImage(const osg::Image* image, float alphaThreshold)
+{
+    if ( !hasAlphaChannel(image) )
+        return false;
+
+    PixelReader read(image);
+    for(unsigned t=0; t<(unsigned)image->t(); ++t) 
+    {
+        for(unsigned s=0; s<(unsigned)image->s(); ++s)
+        {
+            osg::Vec4 color = read(s, t);
+            if ( color.a() > alphaThreshold )
+                return false;
+        }
+    }
+    return true;    
+}
+
+
+osg::Image*
+ImageUtils::createOnePixelImage(const osg::Vec4& color)
 {
     osg::Image* image = new osg::Image;
     image->allocateImage(1,1,1, GL_RGBA, GL_UNSIGNED_BYTE);
-    unsigned char *data = image->data(0,0);
-    memset(data, 0, 4);
+    image->setInternalTextureFormat( GL_RGB8A_INTERNAL );
+    PixelWriter write(image);
+    write(color, 0, 0);
     return image;
 }
 
+bool
+ImageUtils::isSingleColorImage(const osg::Image* image, float threshold)
+{
+    PixelReader read(image);
+
+    osg::Vec4 referenceColor = read(0, 0);
+    float refR = referenceColor.r();
+    float refG = referenceColor.g();
+    float refB = referenceColor.b();
+    float refA = referenceColor.a();
+
+    for(unsigned t=0; t<(unsigned)image->t(); ++t) 
+    {
+        for(unsigned s=0; s<(unsigned)image->s(); ++s)
+        {
+            osg::Vec4 color = read(s, t);
+            if (   (fabs(color.r()-refR) > threshold)
+                || (fabs(color.g()-refG) > threshold)
+                || (fabs(color.b()-refB) > threshold)
+                || (fabs(color.a()-refA) > threshold) )
+            {
+                return false;
+            }
+        }
+    }
+    return true;    
+}
+
 bool
 ImageUtils::canConvert( const osg::Image* image, GLenum pixelFormat, GLenum dataType )
 {
@@ -416,10 +531,15 @@ ImageUtils::convert(const osg::Image* image, GLenum pixelFormat, GLenum dataType
 {
     if ( !image )
         return 0L;
-    
-    if ( image->getPixelFormat() == pixelFormat && image->getDataType() == dataType )
-        return cloneImage(image);
 
+    if ( image->getPixelFormat() == pixelFormat && image->getDataType() == dataType)
+    {
+        GLenum texFormat = image->getInternalTextureFormat();
+        if (dataType != GL_UNSIGNED_BYTE
+            || (pixelFormat == GL_RGB  && texFormat == GL_RGB8_INTERNAL)
+            || (pixelFormat == GL_RGBA && texFormat == GL_RGB8A_INTERNAL))
+        return cloneImage(image);
+    }
     if ( !canConvert(image, pixelFormat, dataType) )
         return 0L;
 
@@ -427,9 +547,9 @@ ImageUtils::convert(const osg::Image* image, GLenum pixelFormat, GLenum dataType
     result->allocateImage(image->s(), image->t(), image->r(), pixelFormat, dataType);
 
     if ( pixelFormat == GL_RGB && dataType == GL_UNSIGNED_BYTE )
-        result->setInternalTextureFormat( GL_RGB8 );
+        result->setInternalTextureFormat( GL_RGB8_INTERNAL );
     else if ( pixelFormat == GL_RGBA && dataType == GL_UNSIGNED_BYTE )
-        result->setInternalTextureFormat( GL_RGBA8 );
+        result->setInternalTextureFormat( GL_RGB8A_INTERNAL );
     else
         result->setInternalTextureFormat( pixelFormat );
 
@@ -453,29 +573,29 @@ ImageUtils::convertToRGBA8(const osg::Image* image)
 bool 
 ImageUtils::areEquivalent(const osg::Image *lhs, const osg::Image *rhs)
 {
-	if (lhs == rhs) return true;
+    if (lhs == rhs) return true;
 
-	if ((lhs->s() == rhs->s()) &&
-		(lhs->t() == rhs->t()) &&
-		(lhs->getInternalTextureFormat() == rhs->getInternalTextureFormat()) &&
-		(lhs->getPixelFormat() == rhs->getPixelFormat()) &&
-		(lhs->getDataType() == rhs->getDataType()) &&
-		(lhs->getPacking() == rhs->getPacking()) &&
-		(lhs->getImageSizeInBytes() == rhs->getImageSizeInBytes()))
-	{
-		unsigned int size = lhs->getImageSizeInBytes();
+    if ((lhs->s() == rhs->s()) &&
+        (lhs->t() == rhs->t()) &&
+        (lhs->getInternalTextureFormat() == rhs->getInternalTextureFormat()) &&
+        (lhs->getPixelFormat() == rhs->getPixelFormat()) &&
+        (lhs->getDataType() == rhs->getDataType()) &&
+        (lhs->getPacking() == rhs->getPacking()) &&
+        (lhs->getImageSizeInBytes() == rhs->getImageSizeInBytes()))
+    {
+        unsigned int size = lhs->getImageSizeInBytes();
         const unsigned char* ptr1 = lhs->data();
         const unsigned char* ptr2 = rhs->data();
-		for (unsigned int i = 0; i < size; ++i)
-		{
+        for (unsigned int i = 0; i < size; ++i)
+        {
             if ( *ptr1++ != *ptr2++ )
                 return false;
-		}
+        }
 
         return true;
-	}
+    }
 
-	return false;
+    return false;
 }
 
 bool
@@ -487,6 +607,101 @@ ImageUtils::hasAlphaChannel(const osg::Image* image)
         image->getPixelFormat() == GL_LUMINANCE_ALPHA );
 }
 
+
+bool
+ImageUtils::hasTransparency(const osg::Image* image, float threshold)
+{
+    if ( !image ) return false;
+    PixelReader read(image);
+    for( int t=0; t<image->t(); ++t )
+        for( int s=0; s<image->s(); ++s )
+            if ( read(s, t).a() < threshold )
+                return true;
+    return false;
+}
+
+
+void
+ImageUtils::featherAlphaRegions(osg::Image* image, float maxAlpha)
+{
+    PixelReader read (image);
+    PixelWriter write(image);
+
+    int ns = image->s();
+    int nt = image->t();
+
+    osg::Vec4 n;
+
+    for( int t=0; t<nt; ++t )
+    {
+        bool rowdone = false;
+        for( int s=0; s<ns && !rowdone; ++s )
+        {
+            osg::Vec4 pixel = read(s, t);
+            if ( pixel.a() <= maxAlpha )
+            {
+                bool wrote = false;
+                if ( s < ns-1 ) {
+                    n = read( s+1, t );
+                    if ( n.a() > maxAlpha ) {
+                        write( n, s, t );
+                        wrote = true;
+                    }
+                }
+                if ( !wrote && s > 0 ) {
+                    n = read( s-1, t );
+                    if ( n.a() > maxAlpha ) {
+                        write( n, s, t );
+                        rowdone = true;
+                    }
+                }
+            }
+        }
+    }
+
+    for( int s=0; s<ns; ++s )
+    {
+        bool coldone = false;
+        for( int t=0; t<nt && !coldone; ++t )
+        {
+            osg::Vec4 pixel = read(s, t);
+            if ( pixel.a() <= maxAlpha )
+            {
+                bool wrote = false;
+                if ( t < nt-1 ) {
+                    n = read( s, t+1 );
+                    if ( n.a() > maxAlpha ) {
+                        write( n, s, t );
+                        wrote = true;
+                    }
+                }
+                if ( !wrote && t > 0 ) {
+                    n = read( s, t-1 );
+                    if ( n.a() > maxAlpha ) {
+                        write( n, s, t );
+                        coldone = true;
+                    }
+                }
+            }
+        }
+    }
+}
+
+
+void
+ImageUtils::convertToPremultipliedAlpha(osg::Image* image)
+{
+    PixelReader read(image);
+    PixelWriter write(image);
+    for(int s=0; s<image->s(); ++s) {
+        for( int t=0; t<image->t(); ++t ) {
+            osg::Vec4f c = read(s, t);
+            write( osg::Vec4f(c.r()*c.a(), c.g()*c.a(), c.b()*c.a(), c.a()), s, t);
+        }
+    }
+}
+
+
 bool
 ImageUtils::isCompressed(const osg::Image *image)
 {
