@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -17,56 +17,83 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>
  */
 #include <osgEarth/Map>
+#include <osgEarth/MapFrame>
+#include <osgEarth/MapModelChange>
 #include <osgEarth/Registry>
 #include <osgEarth/TileSource>
-#include <OpenThreads/ScopedLock>
+#include <osgEarth/HeightFieldUtils>
+#include <osgEarth/URI>
 #include <iterator>
 
 using namespace osgEarth;
-using namespace OpenThreads;
 
 #define LC "[Map] "
 
 //------------------------------------------------------------------------
 
-void
-MapCallback::onMapModelChanged( const MapModelChange& change )
+Map::Map( const MapOptions& options ) :
+osg::Referenced      ( true ),
+_mapOptions          ( options ),
+_initMapOptions      ( options ),
+_dataModelRevision   ( 0 )
 {
-    switch( change.getAction() )
+    if (_mapOptions.cachePolicy().isSet() &&
+        _mapOptions.cachePolicy()->usage() == CachePolicy::USAGE_CACHE_ONLY )
     {
-    case MapModelChange::ADD_ELEVATION_LAYER: 
-        onElevationLayerAdded( change.getElevationLayer(), change.getFirstIndex() ); break;
-    case MapModelChange::ADD_IMAGE_LAYER:
-        onImageLayerAdded( change.getImageLayer(), change.getFirstIndex() ); break;
-    case MapModelChange::ADD_MASK_LAYER:
-        onMaskLayerAdded( change.getMaskLayer() ); break;
-    case MapModelChange::ADD_MODEL_LAYER:
-				onModelLayerAdded( change.getModelLayer(), change.getFirstIndex() ); break;
-    case MapModelChange::REMOVE_ELEVATION_LAYER:
-        onElevationLayerRemoved( change.getElevationLayer(), change.getFirstIndex() ); break;
-    case MapModelChange::REMOVE_IMAGE_LAYER:
-        onImageLayerRemoved( change.getImageLayer(), change.getFirstIndex() ); break;
-    case MapModelChange::REMOVE_MASK_LAYER:
-        onMaskLayerRemoved( change.getMaskLayer() ); break;
-    case MapModelChange::REMOVE_MODEL_LAYER:
-        onModelLayerRemoved( change.getModelLayer() ); break;
-    case MapModelChange::MOVE_ELEVATION_LAYER:
-        onElevationLayerMoved( change.getElevationLayer(), change.getFirstIndex(), change.getSecondIndex() ); break;
-    case MapModelChange::MOVE_IMAGE_LAYER:
-        onImageLayerMoved( change.getImageLayer(), change.getFirstIndex(), change.getSecondIndex() ); break;
-		case MapModelChange::MOVE_MODEL_LAYER:
-				onModelLayerMoved( change.getModelLayer(), change.getFirstIndex(), change.getSecondIndex() ); break;
+        OE_INFO << LC << "CACHE-ONLY MODE activated from map" << std::endl;
+    }
+
+    // if the map has a cache policy set, make this the system-wide default, UNLESS
+    // there ALREADY IS a registry default, in which case THAT will override THIS one.
+    // (In other words, whichever one is set first wins. And of course, if the registry
+    // has an override set, that will cancel out all of this.)
+    const optional<CachePolicy> regCachePolicy = Registry::instance()->defaultCachePolicy();
+
+    if ( _mapOptions.cachePolicy().isSet() )
+    {
+        if ( !regCachePolicy.isSet() )
+        {
+            Registry::instance()->setDefaultCachePolicy( *_mapOptions.cachePolicy() );
+            OE_INFO << LC 
+                << "Setting default cache policy from map ("
+                << _mapOptions.cachePolicy()->usageString() << ")" << std::endl;
+        }
+        else
+        {
+            _mapOptions.cachePolicy() = *regCachePolicy;
+            OE_INFO << LC
+                << "Settings map caching policy to default ("
+                << _mapOptions.cachePolicy()->usageString() << ")" << std::endl;
+        }
+    }
+    else if ( regCachePolicy.isSet() )
+    {
+        _mapOptions.cachePolicy() = *regCachePolicy;
+        OE_INFO << LC
+            << "Settings map caching policy to default ("
+            << _mapOptions.cachePolicy()->usageString() << ")" << std::endl;
+    }
+
+    // the map-side dbOptions object holds I/O information for all components.
+    _dbOptions = osg::clone( Registry::instance()->getDefaultOptions() );
+
+    // we do our own caching
+    _dbOptions->setObjectCacheHint( osgDB::Options::CACHE_NONE );
+
+    // store the IO information in the top-level DB Options:
+    _mapOptions.cachePolicy()->apply( _dbOptions.get() );
+    URIContext( _mapOptions.referrer() ).apply( _dbOptions.get() );
+
+    // apply an express tile size if there is one.
+    if ( _mapOptions.elevationTileSize().isSet() )
+    {
+        _elevationLayers.setExpressTileSize( *_mapOptions.elevationTileSize() );
     }
 }
 
-//------------------------------------------------------------------------
-
-Map::Map( const MapOptions& options ) :
-osg::Referenced( true ),
-_mapOptions( options ),
-_dataModelRevision(0)
+Map::~Map()
 {
-    //NOP
+    OE_DEBUG << "~Map" << std::endl;
 }
 
 bool
@@ -77,25 +104,26 @@ Map::isGeocentric() const
         _mapOptions.coordSysType() == MapOptions::CSTYPE_GEOCENTRIC_CUBE;
 }
 
-const osgDB::ReaderWriter::Options*
-Map::getGlobalOptions() const {
+const osgDB::Options*
+Map::getGlobalOptions() const
+{
     return _globalOptions.get();
 }
 
 void
-Map::setGlobalOptions( const osgDB::ReaderWriter::Options* options ) {
+Map::setGlobalOptions( const osgDB::Options* options )
+{
     _globalOptions = options;
 }
 
-int
-Map::getImageLayers( ImageLayerVector& out_list, bool validLayersOnly ) const
+Revision
+Map::getImageLayers( ImageLayerVector& out_list ) const
 {
     out_list.reserve( _imageLayers.size() );
 
     Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
     for( ImageLayerVector::const_iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i )
-        if ( !validLayersOnly || i->get()->getProfile() )
-            out_list.push_back( i->get() );
+        out_list.push_back( i->get() );
 
     return _dataModelRevision;
 }
@@ -137,15 +165,14 @@ Map::getImageLayerAt( int index ) const
         return 0L;
 }
 
-int
-Map::getElevationLayers( ElevationLayerVector& out_list, bool validLayersOnly ) const
+Revision
+Map::getElevationLayers( ElevationLayerVector& out_list ) const
 {
     out_list.reserve( _elevationLayers.size() );
 
     Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
     for( ElevationLayerVector::const_iterator i = _elevationLayers.begin(); i != _elevationLayers.end(); ++i )
-        if ( !validLayersOnly || i->get()->getProfile() )
-            out_list.push_back( i->get() );
+        out_list.push_back( i->get() );
 
     return _dataModelRevision;
 }
@@ -187,19 +214,49 @@ Map::getElevationLayerAt( int index ) const
         return 0L;
 }
 
-int
-Map::getModelLayers( ModelLayerVector& out_list, bool validLayersOnly ) const
+Revision
+Map::getModelLayers( ModelLayerVector& out_list ) const
 {
     out_list.reserve( _modelLayers.size() );
 
     Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
     for( ModelLayerVector::const_iterator i = _modelLayers.begin(); i != _modelLayers.end(); ++i )
-        //if ( !validLayersOnly || i->get()->i->get()->getProfile() )
-            out_list.push_back( i->get() );
+        out_list.push_back( i->get() );
 
     return _dataModelRevision;
 }
 
+ModelLayer*
+Map::getModelLayerByName( const std::string& name ) const
+{
+    Threading::ScopedReadLock( const_cast<Map*>(this)->_mapDataMutex );
+    for( ModelLayerVector::const_iterator i = _modelLayers.begin(); i != _modelLayers.end(); ++i )
+        if ( i->get()->getName() == name )
+            return i->get();
+    return 0L;
+}
+
+ModelLayer*
+Map::getModelLayerByUID( UID layerUID ) const
+{
+    Threading::ScopedReadLock( const_cast<Map*>(this)->_mapDataMutex );
+    for( ModelLayerVector::const_iterator i = _modelLayers.begin(); i != _modelLayers.end(); ++i )
+        if ( i->get()->getUID() == layerUID )
+            return i->get();
+    return 0L;
+}
+
+
+ModelLayer*
+Map::getModelLayerAt( int index ) const
+{
+    Threading::ScopedReadLock( const_cast<Map*>(this)->_mapDataMutex );
+    if ( index >= 0 && index < (int)_modelLayers.size() )
+        return _modelLayers[index].get();
+    else
+        return 0L;
+}
+
 int
 Map::getNumModelLayers() const
 {
@@ -207,9 +264,16 @@ Map::getNumModelLayers() const
     return _modelLayers.size();
 }
 
-MaskLayer*
-Map::getTerrainMaskLayer() const {
-    return _terrainMaskLayer.get();
+int
+Map::getTerrainMaskLayers( MaskLayerVector& out_list ) const
+{
+    out_list.reserve( _terrainMaskLayers.size() );
+
+    Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
+    for( MaskLayerVector::const_iterator i = _terrainMaskLayers.begin(); i != _terrainMaskLayers.end(); ++i )
+        out_list.push_back( i->get() );
+
+    return _dataModelRevision;
 }
 
 void
@@ -235,18 +299,18 @@ Map::getProfile() const
 Cache*
 Map::getCache() const
 {
-    if ( !_cache.valid() && _mapOptions.cache().isSet() )
+    if ( !_cache.valid() )
     {
         Cache* cache = 0L;
+        
+        // if there's a cache in the registry, install it now.
+        if ( Registry::instance()->getCache() )
+        {
+            cache = Registry::instance()->getCache();
+        }
 
-        // if there's a cache override in the registry, install it now.
-	    if ( osgEarth::Registry::instance()->getCacheOverride() )
-	    {
-		    OE_INFO << LC << "Overriding map cache with global cache override" << std::endl;
-		    cache = osgEarth::Registry::instance()->getCacheOverride();
-	    }
-
-        if ( !cache )
+        // or, if a cache is defined in the options, use that.
+        else if ( _mapOptions.cache().isSet() )
         {
             cache = CacheFactory::create( _mapOptions.cache().get() );
         }
@@ -256,7 +320,7 @@ Map::getCache() const
             const_cast<Map*>(this)->setCache( cache );
         }
     }
-	return _cache.get();
+    return _cache.get();
 }
 
 void
@@ -265,17 +329,21 @@ Map::setCache( Cache* cache )
     if (_cache.get() != cache)
     {
         _cache = cache;
-        _cache->setReferenceURI( _mapOptions.referenceURI().value() );
 
-        //Propagate the cache to any of our layers
+        if ( _cache.valid() )
+        {
+            _cache->apply( _dbOptions.get() );
+        }
+
+        // Propagate the cache to any of our layers
         for (ImageLayerVector::iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i)
         {
-            i->get()->setCache( _cache.get() );
+            i->get()->setDBOptions( _dbOptions.get() );
         }
 
         for (ElevationLayerVector::iterator i = _elevationLayers.begin(); i != _elevationLayers.end(); ++i)
         {
-            i->get()->setCache( _cache.get() );
+            i->get()->setDBOptions( _dbOptions.get() );
         }
     }
 }
@@ -287,27 +355,56 @@ Map::addMapCallback( MapCallback* cb ) const
         const_cast<Map*>(this)->_mapCallbacks.push_back( cb );
 }
 
+void 
+Map::removeMapCallback( MapCallback* cb )
+{
+    MapCallbackList::iterator i = std::find( _mapCallbacks.begin(), _mapCallbacks.end(), cb);
+    if (i != _mapCallbacks.end())
+    {
+        _mapCallbacks.erase( i );
+    }
+}
+
+void
+Map::beginUpdate()
+{
+    MapModelChange msg( MapModelChange::BEGIN_BATCH_UPDATE, _dataModelRevision );
+
+    for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
+    {
+        i->get()->onMapModelChanged( msg );
+    }
+}
+
+void
+Map::endUpdate()
+{
+    MapModelChange msg( MapModelChange::END_BATCH_UPDATE, _dataModelRevision );
+ 
+    for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
+    {
+        i->get()->onMapModelChanged( msg );
+    }
+}
+
 void
 Map::addImageLayer( ImageLayer* layer )
 {
+    osgEarth::Registry::instance()->clearBlacklist();
     unsigned int index = -1;
     if ( layer )
     {
-	    //Set options for the map from the layer
-		layer->setReferenceURI( _mapOptions.referenceURI().value() );
+        // Set the DB options for the map from the layer, including the cache policy.
+        layer->setDBOptions( _dbOptions.get() );
 
-        //propagate the cache to the layer:
-        if ( _mapOptions.cache().isSet() && _mapOptions.cache()->cacheOnly().isSetTo( true ) )
-		{
-			layer->setCacheOnly( true );
-		}
-
-		//Set the Cache for the MapLayer to our cache.
-		layer->setCache( this->getCache() );
+        // propagate the cache to the layer:
+        layer->setCache( this->getCache() );
 
         // Tell the layer the map profile, if possible:
         if ( _profile.valid() )
+        {
             layer->setTargetProfileHint( _profile.get() );
+        }
 
         int newRevision;
 
@@ -325,25 +422,19 @@ Map::addImageLayer( ImageLayer* layer )
         {
             i->get()->onMapModelChanged( MapModelChange(
                 MapModelChange::ADD_IMAGE_LAYER, newRevision, layer, index) );
-            //i->get()->onImageLayerAdded( layer, index, newRevision );
-        }	
-    }	
+        }   
+    }   
 }
 
 
 void
 Map::insertImageLayer( ImageLayer* layer, unsigned int index )
 {
+    osgEarth::Registry::instance()->clearBlacklist();
     if ( layer )
     {
         //Set options for the map from the layer
-        layer->setReferenceURI( _mapOptions.referenceURI().value() );
-
-        //propagate the cache to the layer:
-        if ( _mapOptions.cache().isSet() && _mapOptions.cache()->cacheOnly().isSetTo( true ) )
-        {
-            layer->setCacheOnly( true );
-        }
+        layer->setDBOptions( _dbOptions.get() );
 
         //Set the Cache for the MapLayer to our cache.
         layer->setCache( this->getCache() );
@@ -371,27 +462,22 @@ Map::insertImageLayer( ImageLayer* layer, unsigned int index )
         {
             i->get()->onMapModelChanged( MapModelChange(
                 MapModelChange::ADD_IMAGE_LAYER, newRevision, layer, index) );
-        }	
-    }	
+        }   
+    }   
 }
 
 void
 Map::addElevationLayer( ElevationLayer* layer )
 {
+    osgEarth::Registry::instance()->clearBlacklist();
     unsigned int index = -1;
     if ( layer )
     {
-	    //Set options for the map from the layer
-		layer->setReferenceURI( _mapOptions.referenceURI().value() );
+        //Set options for the map from the layer
+        layer->setDBOptions( _dbOptions.get() );
 
-        //propagate the cache to the layer:
-        if ( _mapOptions.cache().isSet() && _mapOptions.cache()->cacheOnly().isSetTo( true ) )
-		{
-			layer->setCacheOnly( true );
-		}
-
-		//Set the Cache for the MapLayer to our cache.
-		layer->setCache( this->getCache() );
+        //Set the Cache for the MapLayer to our cache.
+        layer->setCache( this->getCache() );
         
         // Tell the layer the map profile, if possible:
         if ( _profile.valid() )
@@ -413,13 +499,14 @@ Map::addElevationLayer( ElevationLayer* layer )
         {
             i->get()->onMapModelChanged( MapModelChange(
                 MapModelChange::ADD_ELEVATION_LAYER, newRevision, layer, index) );
-        }	
-    }	
+        }   
+    }   
 }
 
 void 
 Map::removeImageLayer( ImageLayer* layer )
 {
+    osgEarth::Registry::instance()->clearBlacklist();
     unsigned int index = -1;
 
     osg::ref_ptr<ImageLayer> layerToRemove = layer;
@@ -455,6 +542,7 @@ Map::removeImageLayer( ImageLayer* layer )
 void 
 Map::removeElevationLayer( ElevationLayer* layer )
 {
+    osgEarth::Registry::instance()->clearBlacklist();
     unsigned int index = -1;
 
     osg::ref_ptr<ElevationLayer> layerToRemove = layer;
@@ -591,17 +679,18 @@ Map::addModelLayer( ModelLayer* layer )
         {
             Threading::ScopedWriteLock lock( _mapDataMutex );
             _modelLayers.push_back( layer );
-						index = _modelLayers.size() - 1;
+            index = _modelLayers.size() - 1;
             newRevision = ++_dataModelRevision;
         }
 
-        layer->initialize( _mapOptions.referenceURI().get(), this ); //getReferenceURI(), this );        
+        // initialize the model layer
+        layer->initialize( _dbOptions.get() );
 
         // a seprate block b/c we don't need the mutex
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
         {
             i->get()->onMapModelChanged( MapModelChange(
-								MapModelChange::ADD_MODEL_LAYER, newRevision, layer, index ) );
+                MapModelChange::ADD_MODEL_LAYER, newRevision, layer, index ) );
         }
     }
 }
@@ -618,7 +707,8 @@ Map::insertModelLayer( ModelLayer* layer, unsigned int index )
             newRevision = ++_dataModelRevision;
         }
 
-        layer->initialize( _mapOptions.referenceURI().get(), this ); //getReferenceURI(), this );        
+        // initialize the model layer
+        layer->initialize( _dbOptions.get() );
 
         // a seprate block b/c we don't need the mutex
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
@@ -634,6 +724,9 @@ Map::removeModelLayer( ModelLayer* layer )
 {
     if ( layer )
     {
+        //Take a reference to the layer since we will be deleting it
+        osg::ref_ptr< ModelLayer > layerRef = layer;
+
         Revision newRevision;
         {
             Threading::ScopedWriteLock lock( _mapDataMutex );
@@ -651,7 +744,7 @@ Map::removeModelLayer( ModelLayer* layer )
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); ++i )
         {
             i->get()->onMapModelChanged( MapModelChange(
-                MapModelChange::REMOVE_MODEL_LAYER, newRevision, layer) );
+                MapModelChange::REMOVE_MODEL_LAYER, newRevision, layerRef.get()) );
         }
     }
 }
@@ -704,55 +797,122 @@ Map::moveModelLayer( ModelLayer* layer, unsigned int newIndex )
 }
 
 void
-Map::setTerrainMaskLayer( MaskLayer* layer )
+Map::addTerrainMaskLayer( MaskLayer* layer )
 {
     if ( layer )
     {
         Revision newRevision;
         {
             Threading::ScopedWriteLock lock( _mapDataMutex );
-            _terrainMaskLayer = layer;
+            _terrainMaskLayers.push_back(layer);
             newRevision = ++_dataModelRevision;
         }
 
-        layer->initialize( _mapOptions.referenceURI().value(), this ); //getReferenceURI(), this );
+        layer->initialize( _dbOptions.get(), this );
 
         // a separate block b/c we don't need the mutex   
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
         {
             i->get()->onMapModelChanged( MapModelChange(
                 MapModelChange::ADD_MASK_LAYER, newRevision, layer) );
-        }	
-    }
-    else
-    {
-        removeTerrainMaskLayer();
+        }
     }
 }
 
 void
-Map::removeTerrainMaskLayer()
+Map::removeTerrainMaskLayer( MaskLayer* layer )
 {
-    if ( _terrainMaskLayer.valid() )
+    if ( layer )
     {
+        //Take a reference to the layer since we will be deleting it
+        osg::ref_ptr< MaskLayer > layerRef = layer;
         Revision newRevision;
-
-        osg::ref_ptr<MaskLayer> layer = _terrainMaskLayer.get();
         {
             Threading::ScopedWriteLock lock( _mapDataMutex );
-            _terrainMaskLayer = 0L;
-            newRevision = ++_dataModelRevision;
+            for( MaskLayerVector::iterator i = _terrainMaskLayers.begin(); i != _terrainMaskLayers.end(); ++i )
+            {
+                if ( i->get() == layer )
+                {
+                    _terrainMaskLayers.erase( i );
+                    newRevision = ++_dataModelRevision;
+                    break;
+                }
+            }
         }
         
         // a separate block b/c we don't need the mutex   
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
         {
             i->get()->onMapModelChanged( MapModelChange(
-                MapModelChange::REMOVE_MASK_LAYER, newRevision, layer.get()) );
-        }	
+                MapModelChange::REMOVE_MASK_LAYER, newRevision, layerRef.get()) );
+        }   
     }
 }
 
+
+void
+Map::clear()
+{
+    ImageLayerVector     imageLayersRemoved;
+    ElevationLayerVector elevLayersRemoved;
+    ModelLayerVector     modelLayersRemoved;
+    MaskLayerVector      maskLayersRemoved;
+
+    Revision newRevision;
+    {
+        Threading::ScopedWriteLock lock( _mapDataMutex );
+
+        imageLayersRemoved.swap( _imageLayers );
+        elevLayersRemoved.swap ( _elevationLayers );
+        modelLayersRemoved.swap( _modelLayers );
+
+        // Because you cannot remove a mask layer once it's in place
+        //maskLayersRemoved.swap ( _terrainMaskLayers );
+
+        // calculate a new revision.
+        newRevision = ++_dataModelRevision;
+    }
+    
+    // a separate block b/c we don't need the mutex   
+    for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
+    {
+        for( ImageLayerVector::iterator k = imageLayersRemoved.begin(); k != imageLayersRemoved.end(); ++k )
+            i->get()->onMapModelChanged( MapModelChange(MapModelChange::REMOVE_IMAGE_LAYER, newRevision, k->get()) );
+        for( ElevationLayerVector::iterator k = elevLayersRemoved.begin(); k != elevLayersRemoved.end(); ++k )
+            i->get()->onMapModelChanged( MapModelChange(MapModelChange::REMOVE_ELEVATION_LAYER, newRevision, k->get()) );
+        for( ModelLayerVector::iterator k = modelLayersRemoved.begin(); k != modelLayersRemoved.end(); ++k )
+            i->get()->onMapModelChanged( MapModelChange(MapModelChange::REMOVE_MODEL_LAYER, newRevision, k->get()) );
+        //for( MaskLayerVector::iterator k = maskLayersRemoved.begin(); k != maskLayersRemoved.end(); ++k )
+        //    i->get()->onMapModelChanged( MapModelChange(MapModelChange::REMOVE_MASK_LAYER, newRevision, k->get()) );
+    }
+}
+
+
+void
+Map::setLayersFromMap( const Map* map )
+{
+    this->clear();
+
+    if ( map )
+    {
+        ImageLayerVector newImages;
+        map->getImageLayers( newImages );
+        for( ImageLayerVector::iterator i = newImages.begin(); i != newImages.end(); ++i )
+            addImageLayer( i->get() );
+
+        ElevationLayerVector newElev;
+        map->getElevationLayers( newElev );
+        for( ElevationLayerVector::iterator i = newElev.begin(); i != newElev.end(); ++i )
+            addElevationLayer( i->get() );
+
+        ModelLayerVector newModels;
+        map->getModelLayers( newModels );
+        for( ModelLayerVector::iterator i = newModels.begin(); i != newModels.end(); ++i )
+            addModelLayer( i->get() );
+    }
+}
+
+
 void
 Map::calculateProfile()
 {
@@ -826,6 +986,15 @@ Map::calculateProfile()
             }
         }
 
+        // convert the profile to Plate Carre if necessary.
+        if (_profile.valid() &&
+            _profile->getSRS()->isGeographic() && 
+            getMapOptions().coordSysType() == MapOptions::CSTYPE_PROJECTED )
+        {
+            OE_INFO << LC << "Projected display with geographic SRS; activating Plate Carre mode" << std::endl;
+            _profile = _profile->overrideSRS( _profile->getSRS()->createPlateCarreGeographicSRS() );
+        }
+
         // finally, fire an event if the profile has been set.
         if ( _profile.valid() )
         {
@@ -861,238 +1030,44 @@ Map::calculateProfile()
                 layer->setTargetProfileHint( _profile.get() );
             }
         }
-    }
-}
 
-namespace
-{
-    bool
-    s_getHeightField(const TileKey& key,
-                     const ElevationLayerVector& elevLayers,
-                     const Profile* mapProfile,
-                     bool fallback,
-                     ElevationInterpolation interpolation,
-                     ElevationSamplePolicy samplePolicy,
-                     osg::ref_ptr<osg::HeightField>& out_result,
-                     ProgressCallback* progress) 
-    {
-        unsigned int lowestLOD = key.getLevelOfDetail();
-        bool hfInitialized = false;
-
-        typedef std::map< TerrainLayer*, bool > LayerValidMap;
-        LayerValidMap layerValidMap;
-
-	    //Get a HeightField for each of the enabled layers
-	    GeoHeightFieldVector heightFields;
-
-        unsigned int numValidHeightFields = 0;
-
-        
-        //First pass:  Try to get the exact LOD requested for each enabled heightfield
-        for( ElevationLayerVector::const_iterator i = elevLayers.begin(); i != elevLayers.end(); i++ )
+        // create a "proxy" profile to use when querying elevation layers with a vertical datum
+        if ( _profile->getSRS()->getVerticalDatum() != 0L )
         {
-            ElevationLayer* layer = i->get();
-            if (layer->getProfile() && layer->getEnabled() )
-            {
-                osg::HeightField* hf = layer->createHeightField( key, progress );
-                //osg::ref_ptr< osg::HeightField > hf;
-                //layer->getHeightField( key, hf, progress );
-                layerValidMap[ layer ] = (hf != 0L); //hf.valid();
-                if ( hf )
-                //if (hf.valid())
-                {
-                    numValidHeightFields++;
-                    GeoHeightField ghf( hf, key.getExtent(), layer->getProfile()->getVerticalSRS() );
-                    heightFields.push_back( ghf );
-                }
-            }
+            ProfileOptions po = _profile->toProfileOptions();
+            po.vsrsString().unset();
+            _profileNoVDatum = Profile::create(po);
         }
-
-        //If we didn't get any heightfields and weren't requested to fallback, just return NULL
-        if (numValidHeightFields == 0 && !fallback)
+        else
         {
-            return false;
+            _profileNoVDatum = _profile;
         }
-
-        //Second pass:  We were either asked to fallback or we might have some heightfields at the requested LOD and some that are NULL
-        //              Fall back on parent tiles to fill in the missing data if possible.
-        for( ElevationLayerVector::const_iterator i = elevLayers.begin(); i != elevLayers.end(); i++ )
-        {
-            ElevationLayer* layer = i->get();
-
-            if (layer->getProfile() && layer->getEnabled() )
-            {
-                if (!layerValidMap[ layer ])
-                {
-                    TileKey hf_key = key;
-                    osg::ref_ptr< osg::HeightField > hf;
-                    while (hf_key.valid())
-                    {
-                        hf = layer->createHeightField( hf_key, progress );
-                        if ( hf.valid() )
-                            break;
-
-                        hf_key = hf_key.createParentKey();
-                    }
-
-                    if (hf.valid())
-                    {
-                        if ( hf_key.getLevelOfDetail() < lowestLOD )
-                            lowestLOD = hf_key.getLevelOfDetail();
-
-                        heightFields.push_back( GeoHeightField(
-                            hf.get(), hf_key.getExtent(), layer->getProfile()->getVerticalSRS() ) );
-                    }
-                }
-            }
-        }
-
-	    if (heightFields.size() == 0)
-	    {
-	        //If we got no heightfields, return NULL
-		    return false;
-	    }
-
-	    else if (heightFields.size() == 1)
-	    {
-            if ( lowestLOD == key.getLevelOfDetail() )
-            {
-		        //If we only have on heightfield, just return it.
-		        out_result = heightFields[0].takeHeightField();
-            }
-            else
-            {
-                GeoHeightField geoHF = heightFields[0].createSubSample( key.getExtent(), interpolation);
-                out_result = geoHF.takeHeightField();
-                hfInitialized = true;
-            }
-	    }
-
-	    else
-	    {
-		    //If we have multiple heightfields, we need to composite them together.
-		    unsigned int width = 0;
-		    unsigned int height = 0;
-
-		    for (GeoHeightFieldVector::const_iterator i = heightFields.begin(); i < heightFields.end(); ++i)
-		    {
-			    if (i->getHeightField()->getNumColumns() > width) 
-                    width = i->getHeightField()->getNumColumns();
-			    if (i->getHeightField()->getNumRows() > height) 
-                    height = i->getHeightField()->getNumRows();
-		    }
-		    out_result = new osg::HeightField();
-		    out_result->allocate( width, height );
-
-		    //Go ahead and set up the heightfield so we don't have to worry about it later
-            double minx, miny, maxx, maxy;
-            key.getExtent().getBounds(minx, miny, maxx, maxy);
-            double dx = (maxx - minx)/(double)(out_result->getNumColumns()-1);
-            double dy = (maxy - miny)/(double)(out_result->getNumRows()-1);
-
-            const VerticalSpatialReference* vsrs = mapProfile->getVerticalSRS();
-            
-		    //Create the new heightfield by sampling all of them.
-            for (unsigned int c = 0; c < width; ++c)
-            {
-                double geoX = minx + (dx * (double)c);
-                for (unsigned r = 0; r < height; ++r)
-                {
-                    double geoY = miny + (dy * (double)r);
-
-                    //Collect elevations from all of the layers
-                    std::vector<float> elevations;
-                    for (GeoHeightFieldVector::iterator itr = heightFields.begin(); itr != heightFields.end(); ++itr)
-                    {
-                        const GeoHeightField& geoHF = *itr;
-
-                        float elevation = 0.0f;
-                        if ( geoHF.getElevation(key.getExtent().getSRS(), geoX, geoY, interpolation, vsrs, elevation) )
-                        {
-                            if (elevation != NO_DATA_VALUE)
-                            {
-                                elevations.push_back(elevation);
-                            }
-                        }
-                    }
-
-                    float elevation = NO_DATA_VALUE;
-
-                    //The list of elevations only contains valid values
-                    if (elevations.size() > 0)
-                    {
-                        if (samplePolicy == SAMPLE_FIRST_VALID)
-                        {
-                            elevation = elevations[0];
-                        }
-                        else if (samplePolicy == SAMPLE_HIGHEST)
-                        {
-                            elevation = -FLT_MAX;
-                            for (unsigned int i = 0; i < elevations.size(); ++i)
-                            {
-                                if (elevation < elevations[i]) elevation = elevations[i];
-                            }
-                        }
-                        else if (samplePolicy == SAMPLE_LOWEST)
-                        {
-                            elevation = FLT_MAX;
-                            for (unsigned i = 0; i < elevations.size(); ++i)
-                            {
-                                if (elevation > elevations[i]) elevation = elevations[i];
-                            }
-                        }
-                        else if (samplePolicy == SAMPLE_AVERAGE)
-                        {
-                            elevation = 0.0;
-                            for (unsigned i = 0; i < elevations.size(); ++i)
-                            {
-                                elevation += elevations[i];
-                            }
-                            elevation /= (float)elevations.size();
-                        }
-                    }
-                    out_result->setHeight(c, r, elevation);
-                }
-            }
-	    }
-
-	    //Replace any NoData areas with 0
-	    if (out_result.valid())
-	    {
-		    ReplaceInvalidDataOperator o;
-		    o.setValidDataOperator(new osgTerrain::NoDataValue(NO_DATA_VALUE));
-		    o( out_result.get() );
-	    }
-
-	    //Initialize the HF values for osgTerrain
-	    if (out_result.valid() && !hfInitialized )
-	    {	
-		    //Go ahead and set up the heightfield so we don't have to worry about it later
-		    double minx, miny, maxx, maxy;
-		    key.getExtent().getBounds(minx, miny, maxx, maxy);
-		    out_result->setOrigin( osg::Vec3d( minx, miny, 0.0 ) );
-		    double dx = (maxx - minx)/(double)(out_result->getNumColumns()-1);
-		    double dy = (maxy - miny)/(double)(out_result->getNumRows()-1);
-		    out_result->setXInterval( dx );
-		    out_result->setYInterval( dy );
-		    out_result->setBorderWidth( 0 );
-	    }
-
-	    return out_result.valid();
     }
 }
 
 
 bool
-Map::getHeightField(const TileKey& key,
-                    bool fallback,
+Map::getHeightField(const TileKey&                  key,
+                    bool                            fallback,
                     osg::ref_ptr<osg::HeightField>& out_result,
-                    ElevationInterpolation interpolation,
-                    ElevationSamplePolicy samplePolicy,
-                    ProgressCallback* progress) const
+                    bool*                           out_isFallback,
+                    bool                            convertToHAE,
+                    ElevationSamplePolicy           samplePolicy,
+                    ProgressCallback*               progress) const
 {
     Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
-    return s_getHeightField( key, _elevationLayers, getProfile(), fallback, interpolation, samplePolicy, out_result, progress );
+
+    ElevationInterpolation interp = getMapOptions().elevationInterpolation().get();    
+
+    return _elevationLayers.createHeightField(
+        key, 
+        fallback, 
+        convertToHAE ? _profileNoVDatum.get() : 0L,
+        interp, 
+        samplePolicy, 
+        out_result,  
+        out_isFallback,
+        progress );
 }
 
 bool
@@ -1110,29 +1085,18 @@ Map::sync( MapFrame& frame ) const
             if ( !frame._initialized )
                 frame._imageLayers.reserve( _imageLayers.size() );
             frame._imageLayers.clear();
-            if ( frame._copyValidDataOnly )
-            {
-                for( ImageLayerVector::const_iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i )
-                    if ( i->get()->getProfile() )
-                        frame._imageLayers.push_back( i->get() );
-            }
-            else
-                std::copy( _imageLayers.begin(), _imageLayers.end(), std::back_inserter(frame._imageLayers) );
+            std::copy( _imageLayers.begin(), _imageLayers.end(), std::back_inserter(frame._imageLayers) );
         }
 
         if ( frame._parts & ELEVATION_LAYERS )
         {
-            if ( !frame._initialized )
-                frame._elevationLayers.reserve( _elevationLayers.size() );
-            frame._elevationLayers.clear();
-            if ( frame._copyValidDataOnly )
-            {
-                for( ElevationLayerVector::const_iterator i = _elevationLayers.begin(); i != _elevationLayers.end(); ++i )
-                    if ( i->get()->getProfile() )
-                        frame._elevationLayers.push_back( i->get() );
-            }
-            else
-                std::copy( _elevationLayers.begin(), _elevationLayers.end(), std::back_inserter(frame._elevationLayers) );
+            frame._elevationLayers = _elevationLayers;
+            //if ( !frame._initialized )
+            //    frame._elevationLayers.reserve( _elevationLayers.size() );
+            //frame._elevationLayers.clear();
+            //std::copy( _elevationLayers.begin(), _elevationLayers.end(), std::back_inserter(frame._elevationLayers) );
+            if ( _mapOptions.elevationTileSize().isSet() )
+                frame._elevationLayers.setExpressTileSize( *_mapOptions.elevationTileSize() );
         }
 
         if ( frame._parts & MODEL_LAYERS )
@@ -1143,6 +1107,14 @@ Map::sync( MapFrame& frame ) const
             std::copy( _modelLayers.begin(), _modelLayers.end(), std::back_inserter(frame._modelLayers) );
         }
 
+        if ( frame._parts & MASK_LAYERS )
+        {
+          if ( !frame._initialized )
+              frame._maskLayers.reserve( _terrainMaskLayers.size() );
+          frame._maskLayers.clear();
+          std::copy( _terrainMaskLayers.begin(), _terrainMaskLayers.end(), std::back_inserter(frame._maskLayers) );
+        }
+
         // sync the revision numbers.
         frame._initialized = true;
         frame._mapDataModelRevision = _dataModelRevision;
@@ -1151,98 +1123,3 @@ Map::sync( MapFrame& frame ) const
     }    
     return result;
 }
-
-//------------------------------------------------------------------------
-
-MapFrame::MapFrame( const Map* map, Map::ModelParts parts, const std::string& name ) :
-_initialized( false ),
-_map( map ),
-_name( name ),
-_mapInfo( map ),
-_parts( parts ),
-_copyValidDataOnly( false )
-{
-    sync();
-}
-
-MapFrame::MapFrame( const Map* map, bool copyValidDataOnly, Map::ModelParts parts, const std::string& name ) :
-_initialized( false ),
-_map( map ),
-_name( name ),
-_mapInfo( map ),
-_parts( parts ),
-_copyValidDataOnly( copyValidDataOnly )
-{
-    sync();
-}
-
-MapFrame::MapFrame( const MapFrame& src, const std::string& name ) :
-_initialized( src._initialized ),
-_map( src._map.get() ),
-_name( name ),
-_mapInfo( src._mapInfo ), // src._map.get() ),
-_parts( src._parts ),
-_copyValidDataOnly( src._copyValidDataOnly ),
-_mapDataModelRevision( src._mapDataModelRevision ),
-_imageLayers( src._imageLayers ),
-_elevationLayers( src._elevationLayers ),
-_modelLayers( src._modelLayers )
-{
-    //no sync required here; we copied the arrays etc
-}
-
-bool
-MapFrame::sync()
-{
-    return _map->sync( *this );
-}
-
-bool
-MapFrame::getHeightField(const TileKey& key,
-                            bool fallback,
-                            osg::ref_ptr<osg::HeightField>& out_hf,
-                            ElevationInterpolation interpolation,
-                            ElevationSamplePolicy samplePolicy,
-                            ProgressCallback* progress) const
-{
-    return s_getHeightField( key, _elevationLayers, _mapInfo.getProfile(), fallback, interpolation, samplePolicy, out_hf, progress );
-}
-
-int
-MapFrame::indexOf( ImageLayer* layer ) const
-{
-    ImageLayerVector::const_iterator i = std::find( _imageLayers.begin(), _imageLayers.end(), layer );
-    return i != _imageLayers.end() ? i - _imageLayers.begin() : -1;
-}
-
-int
-MapFrame::indexOf( ElevationLayer* layer ) const
-{
-    ElevationLayerVector::const_iterator i = std::find( _elevationLayers.begin(), _elevationLayers.end(), layer );
-    return i != _elevationLayers.end() ? i - _elevationLayers.begin() : -1;
-}
-
-int
-MapFrame::indexOf( ModelLayer* layer ) const
-{
-    ModelLayerVector::const_iterator i = std::find( _modelLayers.begin(), _modelLayers.end(), layer );
-    return i != _modelLayers.end() ? i - _modelLayers.begin() : -1;
-}
-
-ImageLayer*
-MapFrame::getImageLayerByUID( UID uid ) const
-{
-    for(ImageLayerVector::const_iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i )
-        if ( i->get()->getUID() == uid )
-            return i->get();
-    return 0L;
-}
-
-ImageLayer*
-MapFrame::getImageLayerByName( const std::string& name ) const
-{
-    for(ImageLayerVector::const_iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i )
-        if ( i->get()->getName() == name )
-            return i->get();
-    return 0L;
-}
