@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -22,6 +22,7 @@
 #include <osgEarth/TileKey>
 #include <osgEarth/Cube>
 #include <osgEarth/SpatialReference>
+#include <osgEarth/StringUtils>
 #include <osgDB/FileNameUtils>
 #include <algorithm>
 #include <sstream>
@@ -33,11 +34,11 @@ using namespace osgEarth;
 //------------------------------------------------------------------------
 
 ProfileOptions::ProfileOptions( const ConfigOptions& options ) :
-ConfigOptions( options ),
-_namedProfile( "" ),
-_srsInitString( "" ),
-_vsrsInitString( "" ),
-_bounds( Bounds() ),
+ConfigOptions      ( options ),
+_namedProfile      ( "" ),
+_srsInitString     ( "" ),
+_vsrsInitString    ( "" ),
+_bounds            ( Bounds() ),
 _numTilesWideAtLod0( 1 ),
 _numTilesHighAtLod0( 1 )
 {
@@ -45,9 +46,9 @@ _numTilesHighAtLod0( 1 )
 }
 
 ProfileOptions::ProfileOptions( const std::string& namedProfile ) :
-_srsInitString( "" ),
-_vsrsInitString( "" ),
-_bounds( Bounds() ),
+_srsInitString     ( "" ),
+_vsrsInitString    ( "" ),
+_bounds            ( Bounds() ),
 _numTilesWideAtLod0( 1 ),
 _numTilesHighAtLod0( 1 )
 {
@@ -61,7 +62,9 @@ ProfileOptions::fromConfig( const Config& conf )
         _namedProfile = conf.value();
 
     conf.getIfSet( "srs", _srsInitString );
-    conf.getIfSet( "vsrs", _vsrsInitString );
+
+    conf.getIfSet( "vdatum", _vsrsInitString );
+    conf.getIfSet( "vsrs", _vsrsInitString ); // back compat
 
     if ( conf.hasValue( "xmin" ) && conf.hasValue( "ymin" ) && conf.hasValue( "xmax" ) && conf.hasValue( "ymax" ) )
     {
@@ -87,7 +90,7 @@ ProfileOptions::getConfig() const
     else
     {
         conf.updateIfSet( "srs", _srsInitString );
-        conf.updateIfSet( "vsrs", _vsrsInitString );
+        conf.updateIfSet( "vdatum", _vsrsInitString );
 
         if ( _bounds.isSet() )
         {
@@ -121,44 +124,58 @@ Profile::create(const std::string& srsInitString,
                 unsigned int numTilesWideAtLod0,
                 unsigned int numTilesHighAtLod0)
 {
-    return new Profile(
-        SpatialReference::create( srsInitString ),
-        xmin, ymin, xmax, ymax,
-        VerticalSpatialReference::create( vsrsInitString ),
-        numTilesWideAtLod0,
-        numTilesHighAtLod0 );
+    osg::ref_ptr<osgEarth::SpatialReference> srs = SpatialReference::create( srsInitString, vsrsInitString );
+    if (srs.valid() == true)
+    {
+        return new Profile(
+            srs.get(),
+            xmin, ymin, xmax, ymax,
+            numTilesWideAtLod0,
+            numTilesHighAtLod0 );
+    }
+
+    OE_WARN << LC << "Failed to create profile; unrecognized SRS: \"" << srsInitString << "\"" << std::endl;
+    return NULL;
 }
 
 const Profile*
 Profile::create(const SpatialReference* srs,
                 double xmin, double ymin, double xmax, double ymax,
-                const VerticalSpatialReference* vsrs,
                 unsigned int numTilesWideAtLod0,
                 unsigned int numTilesHighAtLod0)
 {
-    return new Profile(
-        srs,
-        xmin, ymin, xmax, ymax,
-        vsrs,
-        numTilesWideAtLod0,
-        numTilesHighAtLod0 );
+    if (srs != NULL)
+    {
+        return new Profile(
+            srs,
+            xmin, ymin, xmax, ymax,
+            numTilesWideAtLod0,
+            numTilesHighAtLod0 );
+    }
+
+    OE_WARN << LC << "Failed to create profile; null SRS" << std::endl;
+    return 0L;
 }
 
 const Profile*
 Profile::create(const SpatialReference* srs,
                 double xmin, double ymin, double xmax, double ymax,
                 double geoxmin, double geoymin, double geoxmax, double geoymax,
-                const VerticalSpatialReference* vsrs,
                 unsigned int numTilesWideAtLod0,
                 unsigned int numTilesHighAtLod0)
 {
-    return new Profile(
-        srs,
-        xmin, ymin, xmax, ymax,
-        geoxmin, geoymin, geoxmax, geoymax,
-        vsrs,
-        numTilesWideAtLod0,
-        numTilesHighAtLod0 );
+    if ( srs )
+    {
+        return new Profile(
+            srs,
+            xmin, ymin, xmax, ymax,
+            geoxmin, geoymin, geoxmax, geoymax,
+            numTilesWideAtLod0,
+            numTilesHighAtLod0 );
+    }
+    
+    OE_WARN << LC << "Failed to create profile; null SRS" << std::endl;
+    return 0L;
 }
 
 const Profile*
@@ -167,34 +184,37 @@ Profile::create(const std::string& srsInitString,
                 unsigned int numTilesWideAtLod0,
                 unsigned int numTilesHighAtLod0)
 {
-    const Profile* named = osgEarth::Registry::instance()->getNamedProfile( srsInitString );
-    if ( named )
-        return const_cast<Profile*>( named );
+    if ( vsrsInitString.empty() && numTilesWideAtLod0 == 0 && numTilesHighAtLod0 == 0 )
+    {
+        const Profile* named = osgEarth::Registry::instance()->getNamedProfile( srsInitString );
+        if ( named )
+            return const_cast<Profile*>( named );
+    }
 
-    osg::ref_ptr<const SpatialReference> srs = SpatialReference::create( srsInitString );
-
-    osg::ref_ptr<const VerticalSpatialReference> vsrs = VerticalSpatialReference::create( vsrsInitString );
+    osg::ref_ptr<const SpatialReference> srs = SpatialReference::create( srsInitString, vsrsInitString );
 
     if ( srs.valid() && srs->isGeographic() )
     {
         return new Profile(
             srs.get(),
             -180.0, -90.0, 180.0, 90.0,
-            vsrs.get(),
             numTilesWideAtLod0, numTilesHighAtLod0 );
     }
     else if ( srs.valid() && srs->isMercator() )
     {
         // automatically figure out proper mercator extents:
-        GDAL_SCOPED_LOCK;
-        double e, dummy;
-        srs->getGeographicSRS()->transform( 180.0, 0.0, srs.get(), e, dummy );
-        return Profile::create( srs.get(), -e, -e, e, e, vsrs.get(), numTilesWideAtLod0, numTilesHighAtLod0 );
+        osg::Vec3d point(180.0, 0.0, 0.0);
+        srs->getGeographicSRS()->transform(point, srs.get(), point);
+        double e = point.x();
+        return Profile::create( srs.get(), -e, -e, e, e, numTilesWideAtLod0, numTilesHighAtLod0 );
+    }
+    else if ( srs.valid() )
+    {
+        OE_WARN << LC << "Failed to create profile; you must provide extents with a projected SRS." << std::endl;
     }
     else
     {
-        OE_WARN << LC << "Failed to create profile; SRS spec requires addition information: \"" << srsInitString << 
-            std::endl;
+        OE_WARN << LC << "Failed to create profile; unrecognized SRS: \"" << srsInitString << "\"" << std::endl;
     }
 
     return NULL;
@@ -241,9 +261,20 @@ Profile::create( const ProfileOptions& options )
     // Next try SRS with default extents
     else if ( options.srsString().isSet() )
     {
-        result = Profile::create(
-            options.srsString().value(),
-            options.vsrsString().value() );
+        if ( options.numTilesWideAtLod0().isSet() && options.numTilesHighAtLod0().isSet() )
+        {
+            result = Profile::create(
+                options.srsString().value(),
+                options.vsrsString().value(),
+                options.numTilesWideAtLod0().value(),
+                options.numTilesHighAtLod0().value() );
+        }
+        else
+        {
+            result = Profile::create(
+                options.srsString().value(),
+                options.vsrsString().value() );
+        }
     }
 
     return result;
@@ -254,11 +285,10 @@ Profile::create( const ProfileOptions& options )
 
 Profile::Profile(const SpatialReference* srs,
                  double xmin, double ymin, double xmax, double ymax,
-                 const VerticalSpatialReference* vsrs,
                  unsigned int numTilesWideAtLod0,
                  unsigned int numTilesHighAtLod0) :
-osg::Referenced( true ),
-_vsrs( vsrs )
+
+osg::Referenced( true )
 {
     _extent = GeoExtent( srs, xmin, ymin, xmax, ymax );
 
@@ -270,18 +300,20 @@ _vsrs( vsrs )
         _extent :
         _extent.transform( _extent.getSRS()->getGeographicSRS() );
 
-    if ( !_vsrs.valid() )
-        _vsrs = Registry::instance()->getDefaultVSRS();
+    // make a profile sig (sans srs) and an srs sig for quick comparisons.
+    ProfileOptions temp = toProfileOptions();
+    _fullSignature = Stringify() << std::hex << hashString( temp.getConfig().toJSON() );
+    temp.vsrsString() = "";
+    _horizSignature = Stringify() << std::hex << hashString( temp.getConfig().toJSON() );
 }
 
 Profile::Profile(const SpatialReference* srs,
                  double xmin, double ymin, double xmax, double ymax,
                  double geo_xmin, double geo_ymin, double geo_xmax, double geo_ymax,
-                 const VerticalSpatialReference* vsrs,
                  unsigned int numTilesWideAtLod0,
                  unsigned int numTilesHighAtLod0 ) :
-osg::Referenced( true ),
-_vsrs( vsrs )
+
+osg::Referenced( true )
 {
     _extent = GeoExtent( srs, xmin, ymin, xmax, ymax );
 
@@ -292,8 +324,14 @@ _vsrs( vsrs )
         srs->getGeographicSRS(),
         geo_xmin, geo_ymin, geo_xmax, geo_ymax );
 
-    if ( !_vsrs.valid() )
-        _vsrs = Registry::instance()->getDefaultVSRS();
+    //if ( !_vsrs.valid() )
+    //    _vsrs = Registry::instance()->getDefaultVSRS();
+
+    // make a profile sig (sans srs) and an srs sig for quick comparisons.
+    ProfileOptions temp = toProfileOptions();
+    _fullSignature = Stringify() << std::hex << hashString( temp.getConfig().toJSON() );
+    temp.vsrsString() = "";
+    _horizSignature = Stringify() << std::hex << hashString( temp.getConfig().toJSON() );
 }
 
 Profile::ProfileType
@@ -329,28 +367,60 @@ Profile::getLatLongExtent() const {
 std::string
 Profile::toString() const
 {
-    std::stringstream buf;
-    buf << "[srs=" << _extent.getSRS()->getName() << ", min=" << _extent.xMin() << "," << _extent.yMin()
+    const SpatialReference* srs = _extent.getSRS();
+    return Stringify()
+        << std::setprecision(16)
+        << "[srs=" << srs->getName() << ", min=" << _extent.xMin() << "," << _extent.yMin()
         << " max=" << _extent.xMax() << "," << _extent.yMax()
         << " lod0=" << _numTilesWideAtLod0 << "," << _numTilesHighAtLod0
-        << " vsrs=" << ( _vsrs.valid() ? _vsrs->getName() : "default" )
+        << " vdatum=" << (srs->getVerticalDatum() ? srs->getVerticalDatum()->getName() : "geodetic")
         << "]";
-    std::string bufStr;
-	bufStr = buf.str();
-	return bufStr;
+}
+
+ProfileOptions
+Profile::toProfileOptions() const
+{
+    ProfileOptions op;
+    op.srsString() = getSRS()->getHorizInitString();
+    op.vsrsString() = getSRS()->getVertInitString();
+    op.bounds()->xMin() = _extent.xMin();
+    op.bounds()->yMin() = _extent.yMin();
+    op.bounds()->xMax() = _extent.xMax();
+    op.bounds()->yMax() = _extent.yMax();
+    op.numTilesWideAtLod0() = _numTilesWideAtLod0;
+    op.numTilesHighAtLod0() = _numTilesHighAtLod0;
+    return op;
+}
+
+
+Profile*
+Profile::overrideSRS( const SpatialReference* srs ) const
+{
+    return new Profile(
+        srs,
+        _extent.xMin(), _extent.yMin(), _extent.xMax(), _extent.yMax(),
+        _numTilesWideAtLod0, _numTilesHighAtLod0 );
 }
 
 void
 Profile::getRootKeys( std::vector<TileKey>& out_keys ) const
+{
+    getAllKeysAtLOD(0, out_keys);
+}
+
+void
+Profile::getAllKeysAtLOD( unsigned lod, std::vector<TileKey>& out_keys ) const
 {
     out_keys.clear();
 
-    for (unsigned int c = 0; c < _numTilesWideAtLod0; ++c)
+    unsigned tx, ty;
+    getNumTiles( lod, tx, ty );
+
+    for(unsigned c=0; c<tx; ++c)
     {
-        for (unsigned int r = 0; r < _numTilesHighAtLod0; ++r)
+        for(unsigned r=0; r<ty; ++r)
         {
-            //TODO: upgrade to support multi-face profile:
-            out_keys.push_back( TileKey(0, c, r, this) ); // lod, x, y, profile
+            out_keys.push_back( TileKey(lod, c, r, this) );
         }
     }
 }
@@ -384,12 +454,13 @@ Profile::getProfileTypeFromSRS(const std::string& srs_string)
 bool
 Profile::isEquivalentTo( const Profile* rhs ) const
 {
-    return
-        rhs &&
-        _extent.isValid() && rhs->getExtent().isValid() &&
-        _extent == rhs->getExtent() &&
-        _numTilesWideAtLod0 == rhs->_numTilesWideAtLod0 &&
-        _numTilesHighAtLod0 == rhs->_numTilesHighAtLod0;
+    return rhs && getFullSignature() == rhs->getFullSignature();
+}
+
+bool
+Profile::isHorizEquivalentTo( const Profile* rhs ) const
+{
+    return rhs && getHorizSignature() == rhs->getHorizSignature();
 }
 
 void
@@ -455,8 +526,11 @@ Profile::createTileKey( double x, double y, unsigned int level ) const
 }
 
 GeoExtent
-Profile::clampAndTransformExtent( const GeoExtent& input ) const
+Profile::clampAndTransformExtent( const GeoExtent& input, bool* out_clamped ) const
 {
+    if ( out_clamped )
+        *out_clamped = false;
+
     // do the clamping in LAT/LONG.
     const SpatialReference* geo_srs = getSRS()->getGeographicSRS();
 
@@ -466,9 +540,14 @@ Profile::clampAndTransformExtent( const GeoExtent& input ) const
         input :
         input.transform( geo_srs );
 
+    // bail out on a bad transform:
     if ( !gcs_input.isValid() )
         return GeoExtent::INVALID;
 
+    // bail out if the extent's do not intersect at all:
+    if ( !gcs_input.intersects(_latlong_extent, false) )
+        return GeoExtent::INVALID;
+
     // clamp it to the profile's extents:
     GeoExtent clamped_gcs_input = GeoExtent(
         gcs_input.getSRS(),
@@ -477,13 +556,19 @@ Profile::clampAndTransformExtent( const GeoExtent& input ) const
         osg::clampBetween( gcs_input.xMax(), _latlong_extent.xMin(), _latlong_extent.xMax() ),
         osg::clampBetween( gcs_input.yMax(), _latlong_extent.yMin(), _latlong_extent.yMax() ) );
 
+    if ( out_clamped )
+        *out_clamped = (clamped_gcs_input != gcs_input);
+
     // finally, transform the clamped extent into this profile's SRS and return it.
     GeoExtent result =
         clamped_gcs_input.getSRS()->isEquivalentTo( this->getSRS() )?
         clamped_gcs_input :
         clamped_gcs_input.transform( this->getSRS() );
 
-    OE_DEBUG << LC << "clamp&xform: input=" << input.toString() << ", output=" << result.toString() << std::endl;
+    if (result.isValid())
+    {
+        OE_DEBUG << LC << "clamp&xform: input=" << input.toString() << ", output=" << result.toString() << std::endl;
+    }
 
     return result;
 }
@@ -493,7 +578,7 @@ void
 Profile::addIntersectingTiles(const GeoExtent& key_ext, std::vector<TileKey>& out_intersectingKeys) const
 {
     // assume a non-crossing extent here.
-    if ( key_ext.crossesDateLine() )
+    if ( key_ext.crossesAntimeridian() )
     {
         OE_WARN << "Profile::addIntersectingTiles cannot process date-line cross" << std::endl;
         return;
@@ -516,6 +601,7 @@ Profile::addIntersectingTiles(const GeoExtent& key_ext, std::vector<TileKey>& ou
     getTileDimensions(destLOD, destTileWidth, destTileHeight);
 
     //Find the LOD that most closely matches the area of the incoming key without going under.
+#if 0
     while (true)
     {
         currLOD++;
@@ -528,6 +614,20 @@ Profile::addIntersectingTiles(const GeoExtent& key_ext, std::vector<TileKey>& ou
         destTileWidth = w;
         destTileHeight = h;
     }
+#else
+    while( true )
+    {
+        currLOD++;
+        double w, h;
+        getTileDimensions(currLOD, w,h);
+        if ( w < keyWidth || h < keyHeight ) break;
+        //double a = w * h;
+        //if (a < keyArea) break;
+        destLOD = currLOD;
+        destTileWidth = w;
+        destTileHeight = h;
+    }
+#endif
 
     //OE_DEBUG << std::fixed << "  Source Tile: " << key.getLevelOfDetail() << " (" << keyWidth << ", " << keyHeight << ")" << std::endl;
     OE_DEBUG << std::fixed << "  Dest Size: " << destLOD << " (" << destTileWidth << ", " << destTileHeight << ")" << std::endl;
@@ -541,6 +641,13 @@ Profile::addIntersectingTiles(const GeoExtent& key_ext, std::vector<TileKey>& ou
     unsigned int numWide, numHigh;
     getNumTiles(destLOD, numWide, numHigh);
 
+    // bail out if the tiles are out of bounds.
+    if ( tileMinX >= (int)numWide || tileMinY >= (int)numHigh ||
+         tileMaxX < 0 || tileMaxY < 0 )
+    {
+        return;
+    }
+
     tileMinX = osg::clampBetween(tileMinX, 0, (int)numWide-1);
     tileMaxX = osg::clampBetween(tileMaxX, 0, (int)numWide-1);
     tileMinY = osg::clampBetween(tileMinY, 0, (int)numHigh-1);
@@ -592,10 +699,10 @@ Profile::getIntersectingTiles(const GeoExtent& extent, std::vector<TileKey>& out
             return;
     }
 
-    if ( ext.crossesDateLine() )
+    if ( ext.crossesAntimeridian() )
     {
         GeoExtent first, second;
-        if (ext.splitAcrossDateLine( first, second ))
+        if (ext.splitAcrossAntimeridian( first, second ))
         {
             addIntersectingTiles( first, out_intersectingKeys );
             addIntersectingTiles( second, out_intersectingKeys );
@@ -606,3 +713,45 @@ Profile::getIntersectingTiles(const GeoExtent& extent, std::vector<TileKey>& out
         addIntersectingTiles( ext, out_intersectingKeys );
     }
 }
+
+unsigned int
+Profile::getEquivalentLOD( const Profile* profile, unsigned int lod ) const
+{    
+    //If the profiles are equivalent, just use the incoming lod
+    if (profile->isEquivalentTo( this ) ) 
+        return lod;
+
+    double rhsWidth, rhsHeight;
+    profile->getTileDimensions( lod, rhsWidth, rhsHeight );
+
+    // safety catch
+    if ( osg::equivalent(rhsWidth, 0.0) || osg::equivalent(rhsHeight, 0.0) )
+    {
+        OE_WARN << LC << "getEquivalentLOD: zero dimension" << std::endl;
+        return lod;
+    }
+
+    double targetWidth = rhsWidth, targetHeight = rhsHeight;
+
+    if ( !profile->getSRS()->isHorizEquivalentTo(getSRS()) )
+    {
+        targetWidth = profile->getSRS()->transformUnits( rhsWidth, getSRS() );
+        targetHeight = profile->getSRS()->transformUnits( rhsHeight, getSRS() );
+    }
+    
+    int currLOD = 0;
+    int destLOD = currLOD;
+
+    //Find the LOD that most closely matches the area of the incoming key without going under.
+    while( true )
+    {
+        currLOD++;
+        double w, h;
+        getTileDimensions(currLOD, w, h);
+        if ( w < targetWidth || h < targetHeight ) break;
+        //double a = w * h;
+        //if (a < keyArea) break;
+        destLOD = currLOD;
+    }
+    return destLOD;
+}
