@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -19,8 +19,10 @@
 
 #include <osgEarth/Registry>
 #include <osgEarth/TileSource>
-#include <osgEarth/HTTPClient>
 #include <osgEarth/FileUtils>
+#include <osgEarth/ThreadingUtils>
+#include <osgEarth/URI>
+#include <osgEarth/HTTPClient>
 
 #include <osg/Notify>
 #include <osg/io_utils>
@@ -37,6 +39,8 @@
 
 #include "VPBOptions"
 
+#define LC "[VPB] "
+
 using namespace osgEarth;
 using namespace osgEarth::Drivers;
 
@@ -69,11 +73,11 @@ public:
         osgTerrain::TerrainTile* terrainTile = dynamic_cast<osgTerrain::TerrainTile*>(&group);
         if (terrainTile)
         {
-    //        OE_DEBUG<<"VPB: Found terrain tile TileID("<<
-				//TileKey::getLOD(terrainTile->getTileID())<<", "<<
-    //            terrainTile->getTileID().x<<", "<<
-    //            terrainTile->getTileID().y<<")"<<std::endl;
-           
+            OE_DEBUG<<"VPB: Found terrain tile TileID("<<
+                TileKey::getLOD(terrainTile->getTileID())<<", "<<
+                terrainTile->getTileID().x<<", "<<
+                terrainTile->getTileID().y<<")"<<std::endl;
+            
             _terrainTiles.push_back(terrainTile);
         }
         else
@@ -144,12 +148,20 @@ public:
         _options( in_options ),
         //_directory_structure( FLAT_TASK_DIRECTORIES ),
         _profile( osgEarth::Registry::instance()->getGlobalGeodeticProfile() ),
-        _maxNumTilesInCache( 128 )
-        {
-	}
-	
-	void initialize( const std::string& referenceURI)
-	{
+        _maxNumTilesInCache( in_options.terrainTileCacheSize().value() ),
+        _initialized( false )
+    {
+    }
+
+    void initialize( const osgDB::Options* dbOptions )
+    {
+        Threading::ScopedMutexLock lock( _initializeMutex );
+
+        if ( _initialized )
+            return;
+
+        _dbOptions = dbOptions;
+
         unsigned int numTilesWideAtLod0, numTilesHighAtLod0;
         _profile->getNumTiles(0, numTilesWideAtLod0, numTilesHighAtLod0);
 
@@ -158,35 +170,29 @@ public:
 
         if ( !_url.empty() )
         {
-			//If the path doesn't contain a server address, get the full path to the file.
-			if (!osgDB::containsServerAddress(_url))
-			{
-				_url = osgEarth::getFullPath(referenceURI, _url);
-			}
-			
-            osg::ref_ptr<osgDB::ReaderWriter::Options> localOptions = new osgDB::ReaderWriter::Options;
+            osg::ref_ptr<osgDB::Options> localOptions = Registry::instance()->cloneOrCreateOptions();
+
             localOptions->setPluginData("osgearth_vpb Plugin",(void*)(1));
-            //_rootNode = osgDB::readNodeFile( _url, localOptions.get() );
+            ReadResult rc = _url.readNode( localOptions.get() );
 
-            HTTPClient::ResultCode rc = HTTPClient::readNodeFile( _url, _rootNode, localOptions.get() );
-
-            if ( rc == HTTPClient::RESULT_OK && _rootNode.valid() )
+            if ( rc.succeeded() )
             {
+                _rootNode = rc.getNode();
                 _baseNameToUse = _options.baseName().value();
 
-                _path = osgDB::getFilePath(_url);
+                _path = osgDB::getFilePath( *_url );
                 if ( _baseNameToUse.empty() )
-                    _baseNameToUse = osgDB::getStrippedName(_url);
-                _extension = osgDB::getFileExtension(_url);
+                    _baseNameToUse = osgDB::getStrippedName( *_url );
+                _extension = osgDB::getFileExtension( *_url );
                 
-                OE_INFO<<"VPB: Loaded root "<<_url<<", path="<<_path<<" base_name="<<_baseNameToUse<<" extension="<<_extension<<std::endl;
+                OE_INFO << LC << "Loaded root "<< _url.full() <<", path="<<_path<<" base_name="<<_baseNameToUse<<" extension="<<_extension<<std::endl;
                 
-                std::string srs = _profile->getSRS()->getInitString(); //.srs();
+                std::string srs = _profile->getSRS()->getHorizInitString(); //->getInitString(); //.srs();
                 
                 osg::CoordinateSystemNode* csn = dynamic_cast<osg::CoordinateSystemNode*>(_rootNode.get());
                 if (csn)
                 {
-                    OE_INFO<<"VPB: CordinateSystemNode found, coordinate system is : "<<csn->getCoordinateSystem()<<std::endl;
+                    OE_INFO << LC << "CSN found: "<<csn->getCoordinateSystem()<<std::endl;
                     
                     srs = csn->getCoordinateSystem();
                 }
@@ -201,15 +207,15 @@ public:
                     double min_x, max_x, min_y, max_y;
                     ct.getRange(min_x, min_y, max_x, max_y);
 
-                    //OE_DEBUG<<"VPB: range("<<min_x<<", "<<min_y<<", "<<max_x<<", "<<max_y<< ")" <<std::endl;
-					//OE_DEBUG<<"VPB: range("<<osg::RadiansToDegrees(min_x)<<", "<<osg::RadiansToDegrees(min_y)<<", "
-					//	<<osg::RadiansToDegrees(max_x)<<", "<<osg::RadiansToDegrees(max_y)<< ")" <<std::endl;
+                    OE_DEBUG << LC << "range("<<min_x<<", "<<min_y<<", "<<max_x<<", "<<max_y<< ")" <<std::endl;
+                    OE_DEBUG << LC << "range("<<osg::RadiansToDegrees(min_x)<<", "<<osg::RadiansToDegrees(min_y)<<", "
+                        <<osg::RadiansToDegrees(max_x)<<", "<<osg::RadiansToDegrees(max_y)<< ")" <<std::endl;
 
                     srs = locator->getCoordinateSystem();
 
                     double aspectRatio = (max_x-min_x)/(max_y-min_y);
                     
-                    //OE_DEBUG<<"VPB: aspectRatio = "<<aspectRatio<<std::endl;
+                    OE_DEBUG << LC << "aspectRatio = "<<aspectRatio<<std::endl;
 
                     if (aspectRatio>1.0)
                     {
@@ -222,8 +228,8 @@ public:
                         numTilesHighAtLod0 = static_cast<unsigned int>(floor(1.0/aspectRatio+0.499999));
                     }
                     
-                    //OE_DEBUG<<"VPB: computed numTilesWideAtLod0 = "<<numTilesWideAtLod0<<std::endl;
-                    //OE_DEBUG<<"VPB: computed numTilesHightAtLod0 = "<<numTilesHighAtLod0<<std::endl;
+                    OE_DEBUG << LC << "computed numTilesWideAtLod0 = "<<numTilesWideAtLod0<<std::endl;
+                    OE_DEBUG << LC << "computed numTilesHightAtLod0 = "<<numTilesHighAtLod0<<std::endl;
                     
                     //if ( _options.valid() )
                     {
@@ -234,8 +240,8 @@ public:
                             numTilesHighAtLod0 = _options.numTilesHighAtLod0().value();
                     }
 
-                    //OE_DEBUG<<"VPB: final numTilesWideAtLod0 = "<<numTilesWideAtLod0<<std::endl;
-                    //OE_DEBUG<<"VPB: final numTilesHightAtLod0 = "<<numTilesHighAtLod0<<std::endl;
+                    OE_DEBUG << LC << "final numTilesWideAtLod0 = "<<numTilesWideAtLod0<<std::endl;
+                    OE_DEBUG << LC << "final numTilesHightAtLod0 = "<<numTilesHighAtLod0<<std::endl;
                    
                     _profile = osgEarth::Profile::create( 
                         srs,
@@ -251,15 +257,16 @@ public:
             }
             else
             {
-                OE_WARN << "VPB: " << HTTPClient::getResultCodeString(rc) << ": " << _url << std::endl;
-                _url = "";
+                OE_WARN << LC << rc.getResultCodeString() << ": " << *_url << std::endl;
+                _url = URI();
             }
         }
         else 
         {
             OE_WARN<<"VPB: No data referenced "<<std::endl;
         }
-
+        
+        _initialized = true;
     }
     
     std::string createTileName( int level, int tile_x, int tile_y )
@@ -321,12 +328,12 @@ public:
         
 		std::string bufStr;
 		bufStr = buf.str();
-        //OE_DEBUG<<"VPB: VPBDatabase::createTileName(), buf.str()=="<< bufStr <<std::endl;
+        OE_DEBUG<<"VPB: VPBDatabase::createTileName(), buf.str()=="<< bufStr <<std::endl;
         
 		return bufStr;
     }
     
-    osgTerrain::TerrainTile* getTerrainTile( const TileKey& key, ProgressCallback* progress )
+    void getTerrainTile( const TileKey& key, ProgressCallback* progress, osg::ref_ptr<osgTerrain::TerrainTile>& out_tile )
     {
         int level = key.getLevelOfDetail();
         unsigned int tile_x, tile_y;
@@ -339,40 +346,34 @@ public:
 
         osgTerrain::TileID tileID(level, tile_x, tile_y);
 
-        osg::ref_ptr<osgTerrain::TerrainTile> tile = findTile(tileID, false);
-        if (tile.valid()) return tile.get();
-        
-        //OE_INFO<<"Max_x = "<<max_x<<std::endl;
-        //OE_INFO<<"Max_y = "<<max_y<<std::endl;
-
-        //OE_INFO<<"base_name = "<<base_name<<" psl="<<primary_split_level<<" ssl="<<secondary_split_level<<std::endl;
-        //OE_INFO<<"level = "<<level<<", x = "<<tile_x<<", tile_y = "<<tile_y<<std::endl;
-        //OE_INFO<<"tile_name "<<createTileName(level, tile_x, tile_y)<<std::endl;
-        //OE_INFO<<"thread "<<OpenThreads::Thread::CurrentThread()<<std::endl;
+        findTile(tileID, false, out_tile);
+        if (out_tile.valid()) 
+            return;
 
         std::string filename = createTileName(level, tile_x, tile_y);
         
+        bool foundInBlacklist = false;
         {
-            OpenThreads::ScopedLock<OpenThreads::Mutex> lock(_blacklistMutex);
-            if (_blacklistedFilenames.count(filename)==1)
-            {
-                //OE_DEBUG<<"VPB: file has been found in black list : "<<filename<<std::endl;
-                insertTile(tileID, 0);
-                return 0;
-            }
+            Threading::ScopedReadLock sharedLock( _blacklistMutex );
+            foundInBlacklist = _blacklistedFilenames.count(filename) == 1;
         }
-        
+        if ( foundInBlacklist )
+        {
+            OE_DEBUG << LC << "file has been found in black list : "<<filename<<std::endl;
+            insertTile(tileID, 0);
+            return; //return 0;
+        }        
 
-        osg::ref_ptr<osgDB::ReaderWriter::Options> localOptions = new osgDB::ReaderWriter::Options;
+        osg::ref_ptr<osgDB::Options> localOptions = Registry::instance()->cloneOrCreateOptions();
+        CachePolicy::NO_CACHE.apply( localOptions.get() );
         localOptions->setPluginData("osgearth_vpb Plugin",(void*)(1));
 
-
-        //osg::ref_ptr<osg::Node> node = osgDB::readNodeFile(filename, localOptions.get());
-        osg::ref_ptr<osg::Node> node;
-        HTTPClient::ResultCode result = HTTPClient::readNodeFile( filename, node, localOptions.get(), progress );
-        if ( result == HTTPClient::RESULT_OK && node.valid() )
+        ReadResult r = URI(filename).readNode( localOptions.get(), progress );
+        if ( r.succeeded() )
         {
-            //OE_INFO<<"Loaded model "<<filename<<std::endl;
+            osg::Node* node = r.getNode();
+
+            //OE_INFO << LC << "Loaded model "<<filename<<std::endl;
             CollectTiles ct;
             node->accept(ct);
 
@@ -400,6 +401,9 @@ public:
                     
                     tile->setTileID(local_tileID);
                     insertTile(local_tileID, tile);
+
+                    if ( local_tileID == tileID )
+                        out_tile = tile;
                 }
 
             }
@@ -408,20 +412,20 @@ public:
         else
         {
             // in the case of an "unrecoverable" error, black-list the URL for this tile.
-            if ( ! HTTPClient::isRecoverable( result ) )
+            if ( ! HTTPClient::isRecoverable( r.code() ) )
             {
-                //OE_INFO<<"Black listing : "<< filename<< " (" << result << ")" << std::endl;
-                OpenThreads::ScopedLock<OpenThreads::Mutex> lock(_blacklistMutex);
-                _blacklistedFilenames.insert(filename);
+                Threading::ScopedWriteLock exclusiveLock( _blacklistMutex );
+                _blacklistedFilenames.insert( filename );
             }
         }
         
-        return findTile(tileID, false);
+        //TODO: just return it instead...
+        //findTile(tileID, false, out_tile);
     }
     
     void insertTile(const osgTerrain::TileID& tileID, osgTerrain::TerrainTile* tile)
     {
-        OpenThreads::ScopedLock<OpenThreads::Mutex> lock(_tileMapMutex);
+        Threading::ScopedWriteLock exclusiveLock( _tileMapMutex );
 
         if ( _tileMap.find(tileID) == _tileMap.end() )
         {
@@ -435,36 +439,40 @@ public:
                 _tileFIFO.pop_front();
                 _tileMap.erase(tileToRemove);
 
-                //OE_DEBUG<<"VPB: Pruned tileID ("<<TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<")"<<std::endl;
+                OE_DEBUG << LC << "Pruned tileID ("<<TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<")"<<std::endl;
             }
 
-            //OE_DEBUG<<"VPB: insertTile ("
-            //    << TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<") " 
-            //    << " tileFIFO.size()=="<<_tileFIFO.size()<<std::endl;
+            OE_DEBUG << LC << "insertTile ("
+                << TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<") " 
+                << " tileFIFO.size()=="<<_tileFIFO.size()<<std::endl;
         }
         else
         {
-            //OE_DEBUG<<"VPB: insertTile ("
-            //    << TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<") " 
-            //    << " ...already in cache!"<<std::endl;
+            OE_DEBUG << LC << "insertTile ("
+                << TileKey::getLOD(tileID)<<", "<<tileID.x<<", "<<tileID.y<<") " 
+                << " ...already in cache!"<<std::endl;
         }
     }
 
-    osgTerrain::TerrainTile* findTile(const osgTerrain::TileID& tileID, bool insertBlankTileIfNotFound = false)
+    void findTile(const osgTerrain::TileID& tileID, bool insertBlankTileIfNotFound, osg::ref_ptr<osgTerrain::TerrainTile>& out_tile)
     {
+        // read with a shared lock
         {
-            OpenThreads::ScopedLock<OpenThreads::Mutex> lock(_tileMapMutex);
+            Threading::ScopedReadLock sharedLock( _tileMapMutex );
             TileMap::iterator itr = _tileMap.find(tileID);
-            if (itr != _tileMap.end()) return itr->second.get();
+            if (itr != _tileMap.end())
+                out_tile = itr->second.get();
         }
 
-        if (insertBlankTileIfNotFound) insertTile(tileID, 0);
+        // upgrade lock and write:
+        if (insertBlankTileIfNotFound) 
+            insertTile(tileID, 0);
 
-        return 0;
+        //return 0;
     }
 
     const VPBOptions _options;
-    std::string _url;
+    URI _url;
     std::string _path;
     std::string _extension;
 
@@ -477,104 +485,109 @@ public:
     
     typedef std::map<osgTerrain::TileID, osg::ref_ptr<osgTerrain::TerrainTile> > TileMap;
     TileMap _tileMap;
-    OpenThreads::Mutex _tileMapMutex;
+    Threading::ReadWriteMutex _tileMapMutex;
     
     typedef std::list<osgTerrain::TileID> TileIDList;
     TileIDList _tileFIFO;
     
     typedef std::set<std::string> StringSet;
     StringSet _blacklistedFilenames;
-    OpenThreads::Mutex _blacklistMutex;
+    Threading::ReadWriteMutex _blacklistMutex;
+
+    bool _initialized;
+    Threading::Mutex _initializeMutex;
+
+    osg::ref_ptr<const osgDB::Options> _dbOptions;
     
 };
 
 class VPBSource : public TileSource
 {
 public:
-    VPBSource( VPBDatabase* vpbDatabase, const VPBOptions& in_options ) : //const VPBOptions* in_options) :  
-        TileSource(in_options),
-        _vpbDatabase(vpbDatabase),
-        _options( in_options ),
-        _referenceUri()
+    VPBSource( VPBDatabase* vpbDatabase, const VPBOptions& in_options ) : 
+        TileSource   ( in_options  ),
+        _vpbDatabase ( vpbDatabase ),
+        _options     ( in_options  )
     {
         //nop
+     }
+
+    Status initialize( const osgDB::Options* dbOptions )
+    {
+        _dbOptions = Registry::instance()->cloneOrCreateOptions( dbOptions );
+        CachePolicy::NO_CACHE.apply( _dbOptions.get() );
+
+        _vpbDatabase->initialize( _dbOptions.get() );
+
+        if ( !getProfile() )
+        {
+            setProfile(_vpbDatabase->_profile.get());
+        }
+
+        return STATUS_OK;
     }
 
-    void initialize( const std::string& referenceURI, const Profile* overrideProfile)
+    osg::Image* createImage( const TileKey& key, ProgressCallback* progress)
     {
-		_referenceUri = referenceURI;
-		_vpbDatabase->initialize(referenceURI);
-		if ( overrideProfile)
-		{
-			setProfile( overrideProfile );
-		}
-		else
-		{
-			setProfile(_vpbDatabase->_profile.get());
-		}
+        osg::Image * ret = NULL;
+        //TODO:  Make VPB driver use progress callback
+        osg::ref_ptr<osgTerrain::TerrainTile> tile;
+        _vpbDatabase->getTerrainTile(key, progress, tile);
+        if (tile.valid())
+        {        
+            int layerNum = _options.layer().value();
+            const optional<std::string> & layerSetName = _options.layerSetName();
+
+            int numColorLayers = (int)tile->getNumColorLayers();
+            if(layerNum > numColorLayers)
+                layerNum = 0;
+            if (layerNum < numColorLayers)
+            {
+                osgTerrain::Layer* layer = tile->getColorLayer(layerNum);
+
+                osgTerrain::ImageLayer* imageLayer = dynamic_cast<osgTerrain::ImageLayer*>(layer);
+                if (imageLayer)
+                {
+                    OE_DEBUG << LC << "createImage(" << key.str() << " layerNum=" << layerNum << ") successful." <<std::endl;
+                    ret = new osg::Image( *imageLayer->getImage() );
+                }
+                else
+                {
+                    osgTerrain::SwitchLayer* switchLayer = dynamic_cast<osgTerrain::SwitchLayer*>(layer);
+                    if (switchLayer && layerSetName.isSet())
+                    {
+                        for(unsigned int si=0; !imageLayer && si<switchLayer->getNumLayers(); ++si)
+                        {
+                            if(switchLayer->getSetName(si) == layerSetName.value())
+                            {
+                                imageLayer = dynamic_cast<osgTerrain::ImageLayer*>(switchLayer->getLayer(si));
+                            }
+                        }
+                    }
+                    if(imageLayer)
+                    {
+                        OE_DEBUG << LC << "createImage(" << key.str() << " layerSet=" << layerSetName.value() << ") successful." <<std::endl;
+                        ret = new osg::Image( *imageLayer->getImage() );
+                    }
+                }
+            }
+            if(!ret)
+            {
+                OE_DEBUG << LC << "createImage(" << key.str() << " layerSet=" << layerSetName.value() << " layerNum=" << layerNum << "/" << numColorLayers << ") failed." <<std::endl;
+            }
+        }
+        else
+        {
+            OE_DEBUG << LC << "createImage(" << key.str() << ") database retrieval failed." <<std::endl;
+        }
+        return ret;
     }
-    
-	osg::Image* createImage( const TileKey& key,
-		ProgressCallback* progress)
-	{
-		osg::Image * ret = NULL;
-		//TODO:  Make VPB driver use progress callback
-		osg::ref_ptr<osgTerrain::TerrainTile> tile = _vpbDatabase->getTerrainTile(key, progress);                
-		if (tile.valid())
-		{        
-			int layerNum = _options.layer().value();
-			const optional<std::string> & layerSetName = _options.layerSetName();
 
-			int numColorLayers = (int)tile->getNumColorLayers();
-			if(layerNum > numColorLayers)
-				layerNum = 0;
-			if (layerNum < numColorLayers)
-			{
-				osgTerrain::Layer* layer = tile->getColorLayer(layerNum);
-
-				osgTerrain::ImageLayer* imageLayer = dynamic_cast<osgTerrain::ImageLayer*>(layer);
-				if (imageLayer)
-				{
-					//OE_DEBUG<<"VPB: createImage(" << key.str() << " layerNum=" << layerNum << ") successful." <<std::endl;
-					ret = new osg::Image( *imageLayer->getImage() );
-				}
-				else
-				{
-					osgTerrain::SwitchLayer* switchLayer = dynamic_cast<osgTerrain::SwitchLayer*>(layer);
-					if (switchLayer && layerSetName.isSet())
-					{
-						for(unsigned int si=0; !imageLayer && si<switchLayer->getNumLayers(); ++si)
-						{
-							if(switchLayer->getSetName(si) == layerSetName.value())
-							{
-								imageLayer = dynamic_cast<osgTerrain::ImageLayer*>(switchLayer->getLayer(si));
-							}
-						}
-					}
-					if(imageLayer)
-					{
-						//OE_DEBUG<<"VPB: createImage(" << key.str() << " layerSet=" << layerSetName.value() << ") successful." <<std::endl;
-						ret = new osg::Image( *imageLayer->getImage() );
-					}
-				}
-			}
-			if(!ret)
-			{
-				//OE_DEBUG<<"VPB: createImage(" << key.str() << " layerSet=" << layerSetName.value() << " layerNum=" << layerNum << "/" << numColorLayers << ") failed." <<std::endl;
-			}
-		}
-		else
-		{
-			//OE_DEBUG<<"VPB: createImage(" << key.str() << ") database retrieval failed." <<std::endl;
-		}
-		return ret;
-	}
-
-    osg::HeightField* createHeightField( const TileKey& key,
-                                         ProgressCallback* progress
-                                         )
+    osg::HeightField* createHeightField( const TileKey&        key,
+                                         ProgressCallback*     progress )
     {
-        osg::ref_ptr<osgTerrain::TerrainTile> tile = _vpbDatabase->getTerrainTile(key, progress);                
+        osg::ref_ptr<osgTerrain::TerrainTile> tile;
+        _vpbDatabase->getTerrainTile(key, progress, tile);
         if (tile.valid())
         {        
             osgTerrain::Layer* elevationLayer = tile->getElevationLayer();
@@ -596,9 +609,9 @@ public:
     }
 
 private:
-    osg::ref_ptr<VPBDatabase> _vpbDatabase;
-    const VPBOptions _options;
-	std::string	_referenceUri;
+    osg::ref_ptr<VPBDatabase>    _vpbDatabase;
+    const VPBOptions             _options;
+    osg::ref_ptr<osgDB::Options> _dbOptions;
 };
 
 
@@ -622,11 +635,11 @@ class VPBSourceFactory : public TileSourceDriver
 
             VPBOptions vpbOptions( getTileSourceOptions(options) );
 
-            std::string url = vpbOptions.url().value();
+            URI url = vpbOptions.url().value();
             if ( !url.empty() )
             {                
                 OpenThreads::ScopedLock<OpenThreads::Mutex> lock(vpbDatabaseMapMutex);
-                osg::observer_ptr<VPBDatabase>& db_ptr = vpbDatabaseMap[url]; //get or create
+                osg::observer_ptr<VPBDatabase>& db_ptr = vpbDatabaseMap[*url]; //get or create
                 
                 if (!db_ptr) db_ptr = new VPBDatabase( vpbOptions );
                 
