@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -20,6 +20,7 @@
 #include <osgEarth/TileSource>
 #include <osgEarth/ImageToHeightFieldConverter>
 #include <osgEarth/Registry>
+#include <osgEarth/TimeControl>
 #include <osgEarth/XmlUtils>
 #include <osgEarthUtil/WMS>
 #include <osgDB/FileNameUtils>
@@ -37,72 +38,94 @@
 #include "TileService"
 #include "WMSOptions"
 
+#define LC "[WMS] "
+
 using namespace osgEarth;
 using namespace osgEarth::Util;
 using namespace osgEarth::Drivers;
 
-// All looping ImageSequences deriving from this class will by in sync due to
-// a shared reference time.
-class SyncImageSequence : public osg::ImageSequence {
-public:
-    SyncImageSequence() { }
+//----------------------------------------------------------------------------
 
-    virtual void update(osg::NodeVisitor* nv) {
-        setReferenceTime( 0.0 );
-        osg::ImageSequence::update( nv );
-    }
-};
+namespace
+{
+    // All looping ImageSequences deriving from this class will be in sync due to
+    // a shared reference time.
+    struct SyncImageSequence : public osg::ImageSequence
+    {
+        SyncImageSequence() : osg::ImageSequence() { }
 
+        virtual void update(osg::NodeVisitor* nv)
+        {
+            setReferenceTime( 0.0 );
+            osg::ImageSequence::update( nv );
+        }
+    };
+}
 
-class WMSSource : public TileSource
+//----------------------------------------------------------------------------
+
+class WMSSource : public TileSource, public SequenceControl
 {
 public:
 	WMSSource( const TileSourceOptions& options ) : TileSource( options ), _options(options)
     {
+        _isPlaying     = false;
+
         if ( _options.times().isSet() )
         {
-            osgEarth::split( _options.times().value(), ",", _timesVec, false );
+            StringTokenizer( *_options.times(), _timesVec, ",", "", false, true );
+            OE_INFO << LC << "WMS-T: found " << _timesVec.size() << " times." << std::endl;
+
+            for( unsigned i=0; i<_timesVec.size(); ++i )
+            {
+                _seqFrameInfoVec.push_back(SequenceFrameInfo());
+                _seqFrameInfoVec.back().timeIdentifier = _timesVec[i];
+            }
         }
 
         // localize it since we might override them:
         _formatToUse = _options.format().value();
-        _srsToUse = _options.srs().value();
+        _srsToUse = _options.wmsVersion().value() == "1.3.0" ? _options.crs().value() : _options.srs().value();
+        if (_srsToUse.empty())
+        {
+            //If they didn't specify a CRS, see if they specified an SRS and try to use that
+            _srsToUse = _options.srs().value();
+        }
     }
 
     /** override */
-    void initialize( const std::string& referenceURI, const Profile* overrideProfile)
+    Status initialize( const osgDB::Options* dbOptions )
     {
         osg::ref_ptr<const Profile> result;
 
-        char sep = _options.url()->find_first_of('?') == std::string::npos? '?' : '&';
+        char sep = _options.url()->full().find_first_of('?') == std::string::npos? '?' : '&';
 
-        std::string capUrl = _options.capabilitiesUrl().value();
+        URI capUrl = _options.capabilitiesUrl().value();
         if ( capUrl.empty() )
         {
-            capUrl = 
-                _options.url().value() + 
+            capUrl = URI(
+                _options.url()->full() + 
                 sep + 
-                "SERVICE=WMS" +
-                "&VERSION=" + _options.wmsVersion().value() +
-                "&REQUEST=GetCapabilities";
+                std::string("SERVICE=WMS") +
+                std::string("&VERSION=") + _options.wmsVersion().value() +
+                std::string("&REQUEST=GetCapabilities") );
         }
 
         //Try to read the WMS capabilities
-        osg::ref_ptr<WMSCapabilities> capabilities = WMSCapabilitiesReader::read( capUrl, 0L ); //getOptions() );
+        osg::ref_ptr<WMSCapabilities> capabilities = WMSCapabilitiesReader::read( capUrl.full(), dbOptions );
         if ( !capabilities.valid() )
         {
-            OE_WARN << "[osgEarth::WMS] Unable to read WMS GetCapabilities." << std::endl;
-            //return;
+            return Status::Error( "Unable to read WMS GetCapabilities." );
         }
         else
         {
-            OE_INFO << "[osgEarth::WMS] Got capabilities from " << capUrl << std::endl;
+            OE_INFO << "[osgEarth::WMS] Got capabilities from " << capUrl.full() << std::endl;
         }
 
         if ( _formatToUse.empty() && capabilities.valid() )
         {
             _formatToUse = capabilities->suggestExtension();
-            OE_NOTICE << "[osgEarth::WMS] No format specified, capabilities suggested extension " << _formatToUse << std::endl;
+            OE_INFO << "[osgEarth::WMS] No format specified, capabilities suggested extension " << _formatToUse << std::endl;
         }
 
         if ( _formatToUse.empty() )
@@ -118,14 +141,14 @@ public:
 
         // first the mandatory keys:
         buf
-            << std::fixed << _options.url().value() << sep
-            << "SERVICE=WMS"
+            << std::fixed << _options.url()->full() << sep
+	    << "SERVICE=WMS"
             << "&VERSION=" << _options.wmsVersion().value()
             << "&REQUEST=GetMap"
             << "&LAYERS=" << _options.layers().value()
             << "&FORMAT=" << ( wmsFormatToUse.empty() ? std::string("image/") + _formatToUse : wmsFormatToUse )
             << "&STYLES=" << _options.style().value()
-            << "&SRS=" << _srsToUse
+            << (_options.wmsVersion().value() == "1.3.0" ? "&CRS=" : "&SRS=") << _srsToUse            
             << "&WIDTH="<< _options.tileSize().value()
             << "&HEIGHT="<< _options.tileSize().value()
             << "&BBOX=%lf,%lf,%lf,%lf";
@@ -135,9 +158,11 @@ public:
             buf << "&TRANSPARENT=" << (_options.transparent() == true ? "TRUE" : "FALSE");
             
 
-		_prototype = "";
+        _prototype = "";
         _prototype = buf.str();
 
+        //OE_NOTICE << "Prototype " << _prototype << std::endl;
+
         osg::ref_ptr<SpatialReference> wms_srs = SpatialReference::create( _srsToUse );
 
         // check for spherical mercator:
@@ -145,10 +170,10 @@ public:
         {
             result = osgEarth::Registry::instance()->getGlobalMercatorProfile();
         }
-		else if (wms_srs.valid() && wms_srs->isEquivalentTo( osgEarth::Registry::instance()->getGlobalGeodeticProfile()->getSRS()))
-		{
-			result = osgEarth::Registry::instance()->getGlobalGeodeticProfile();
-		}
+        else if (wms_srs.valid() && wms_srs->isEquivalentTo( osgEarth::Registry::instance()->getGlobalGeodeticProfile()->getSRS()))
+        {
+            result = osgEarth::Registry::instance()->getGlobalGeodeticProfile();
+        }
 
         // Next, try to glean the extents from the layer list
         if ( capabilities.valid() )
@@ -164,14 +189,14 @@ public:
                 //Check to see if the profile is equivalent to global-geodetic
                 if (wms_srs->isGeographic())
                 {
-					//Try to get the lat lon extents if they are provided
-				    layer->getLatLonExtents(minx, miny, maxx, maxy);
+                    //Try to get the lat lon extents if they are provided
+                    layer->getLatLonExtents(minx, miny, maxx, maxy);
 
-					//If we still don't have any extents, just default to global geodetic.
-					if (!result.valid() && minx == 0 && miny == 0 && maxx == 0 && maxy == 0)
-					{
-						result = osgEarth::Registry::instance()->getGlobalGeodeticProfile();
-					}
+                    //If we still don't have any extents, just default to global geodetic.
+                    if (!result.valid() && minx == 0 && miny == 0 && maxx == 0 && maxy == 0)
+                    {
+                        result = osgEarth::Registry::instance()->getGlobalGeodeticProfile();
+                    }
                 }	
 
                 if (minx == 0 && miny == 0 && maxx == 0 && maxy == 0)
@@ -188,8 +213,8 @@ public:
                 //Add the layer extents to the list of valid areas
                 if (minx != 0 || maxx != 0 || miny != 0 || maxy != 0)
                 {
-                    GeoExtent extent( result->getSRS(), minx, miny, maxx, maxy);                    
-                    getDataExtents().push_back( DataExtent(extent, 0, INT_MAX) );
+                    GeoExtent extent( result->getSRS(), minx, miny, maxx, maxy);
+                    getDataExtents().push_back( DataExtent(extent, 0) );
                 }
             }
         }
@@ -199,18 +224,17 @@ public:
         {
             result = osgEarth::Registry::instance()->getGlobalGeodeticProfile();
         }
-        
 
         // JPL uses an experimental interface called TileService -- ping to see if that's what
         // we are trying to read:
-        std::string tsUrl = _options.tileServiceUrl().value();
-        if (tsUrl.empty() )
+        URI tsUrl = _options.tileServiceUrl().value();
+        if ( tsUrl.empty() )
         {
-            tsUrl = _options.url().value() + sep + std::string("request=GetTileService");
+            tsUrl = URI(_options.url()->full() + sep + std::string("request=GetTileService") );
         }
 
-        OE_INFO << "[osgEarth::WMS] Testing for JPL/TileService at " << tsUrl << std::endl;
-        _tileService = TileServiceReader::read(tsUrl, 0L); //getOptions());
+        OE_INFO << "[osgEarth::WMS] Testing for JPL/TileService at " << tsUrl.full() << std::endl;
+        _tileService = TileServiceReader::read(tsUrl.full(), dbOptions);
         if (_tileService.valid())
         {
             OE_INFO << "[osgEarth::WMS] Found JPL/TileService spec" << std::endl;
@@ -227,7 +251,7 @@ public:
             if (patterns.size() > 0)
             {
                 result = _tileService->createProfile( patterns );
-				_prototype = _options.url().value() + sep + patterns[0].getPrototype();
+                _prototype = _options.url()->full() + sep + patterns[0].getPrototype();
             }
         }
         else
@@ -235,13 +259,26 @@ public:
             OE_INFO << "[osgEarth::WMS] No JPL/TileService spec found; assuming standard WMS" << std::endl;
         }
 
-        //TODO: won't need this for OSG 2.9+, b/c of mime-type support
-        _prototype = _prototype + std::string("&.") + _formatToUse;
+        // Use the override profile if one is passed in.
+        if ( getProfile() == 0L )
+        {
+            setProfile( result.get() );
+        }
 
-        // populate the data metadata:
-        // TODO
+        if ( getProfile() )
+        {
+            OE_NOTICE << "[osgEarth::WMS] Profile=" << getProfile()->toString() << std::endl;
 
-		setProfile( result.get() );
+            // set up the cache options properly for a TileSource.
+            _dbOptions = Registry::instance()->cloneOrCreateOptions( dbOptions );
+            CachePolicy::NO_CACHE.apply( _dbOptions.get() );
+
+            return STATUS_OK;
+        }
+        else
+        {
+            return Status::Error( "Unable to establish profile" );
+        }
     }
 
     /* override */
@@ -258,7 +295,7 @@ public:
         const TileKey&     key, 
         const std::string& extraAttrs,
         ProgressCallback*  progress, 
-        HTTPResponse&      out_response )
+        ReadResult&        out_response )
     {
         osgDB::ReaderWriter* result = 0L;
 
@@ -270,27 +307,36 @@ public:
         }
 
 
-        out_response = HTTPClient::get( uri, 0L, progress ); //getOptions(), progress );
+        out_response = URI( uri ).readString( _dbOptions.get(), progress );
 
-        if ( out_response.isOK() )
+        //...
+        //out_response = HTTPClient::get( uri, 0L, progress ); //getOptions(), progress );
+
+        if ( out_response.succeeded() )
         {
-            const std::string& mt = out_response.getMimeType();
+            // get the mime type:
+            std::string mt = out_response.metadata().value( IOMetadata::CONTENT_TYPE );
 
             if ( mt == "application/vnd.ogc.se_xml" || mt == "text/xml" )
             {
+                std::istringstream content( out_response.getString() );
+
                 // an XML result means there was a WMS service exception:
                 Config se;
-                if ( se.loadXML( out_response.getPartStream(0) ) )
+                if ( se.fromXML(content) )
                 {
                     Config ex = se.child("serviceexceptionreport").child("serviceexception");
-                    if ( !ex.empty() ) {
-                        OE_NOTICE << "WMS Service Exception: " << ex.value() << std::endl;
+                    if ( !ex.empty() )
+                    {
+                        OE_NOTICE << "WMS Service Exception: " << ex.toJSON(true) << std::endl;
                     }
-                    else {
-                        OE_NOTICE << "WMS Response: " << se.toString() << std::endl;
+                    else
+                    {
+                        OE_NOTICE << "WMS Response: " << se.toJSON(true) << std::endl;
                     }
                 }
-                else {
+                else
+                {
                     OE_NOTICE << "WMS: unknown error." << std::endl;
                 }
             }
@@ -299,7 +345,8 @@ public:
                 // really ought to use mime-type support here -GW
                 std::string typeExt = mt.substr( mt.find_last_of("/")+1 );
                 result = osgDB::Registry::instance()->getReaderWriterForExtension( typeExt );
-                if ( !result ) {
+                if ( !result )
+                {
                     OE_NOTICE << "WMS: no reader registered; URI=" << createURI(key) << std::endl;
                 }
             }
@@ -321,17 +368,20 @@ public:
         {
             std::string extras;
             if ( _timesVec.size() == 1 )
-                extras = "TIME=" + _timesVec[0];
+                extras = std::string("TIME=") + _timesVec[0];
 
-            HTTPResponse response;
+            ReadResult response;
             osgDB::ReaderWriter* reader = fetchTileAndReader( key, extras, progress, response );
             if ( reader )
             {
-                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( response.getPartStream( 0 ), 0L ); //getOptions() );
-                if ( readResult.error() ) {
+                std::istringstream buf( response.getString() );
+                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( buf, _dbOptions.get() );
+                if ( readResult.error() )
+                {
                     OE_WARN << "WMS: image read failed for " << createURI(key) << std::endl;
                 }
-                else {
+                else
+                {
                     image = readResult.getImage();
                 }
             }
@@ -347,13 +397,16 @@ public:
 
         for( unsigned int r=0; r<_timesVec.size(); ++r )
         {
-            std::string extraAttrs = "TIME=" + _timesVec[r];
-            HTTPResponse response;
+            std::string extraAttrs = std::string("TIME=") + _timesVec[r];
+            ReadResult response;
             osgDB::ReaderWriter* reader = fetchTileAndReader( key, extraAttrs, progress, response );
             if ( reader )
             {
-                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( response.getPartStream( 0 ), 0L ); //getOptions() );
-                if ( readResult.error() ) {
+                std::istringstream buf( response.getString() );
+
+                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( buf, _dbOptions.get() );
+                if ( readResult.error() )
+                {
                     OE_WARN << "WMS: image read failed for " << createURI(key) << std::endl;
                 }
                 else
@@ -381,49 +434,27 @@ public:
 
         return image.release();
     }
-    
-    ///** creates a 3D image from timestamped data. */
-    //osg::Image* createImageSequence( const TileKey& key, ProgressCallback* progress )
-    //{
-    //    osg::ImageSequence* seq = new osg::ImageSequence();
-
-    //    for( int r=0; r<_timesVec.size(); ++r )
-    //    {
-    //        std::string extraAttrs = "TIME=" + _timesVec[r];
-
-    //        std::string uri = createURI(key);
-    //        std::string delim = uri.find("?") == std::string::npos ? "?" : "&";
-    //        uri = uri + delim + extraAttrs;
-    //        uri = uri + "&." + _formatToUse;
-
-    //        seq->addImageFile( uri );
-    //    }
-
-    //    seq->play();
-    //    seq->setLength( (double)_timesVec.size() );
-    //    seq->setLoopingMode( osg::ImageStream::LOOPING );
-    //    
-    //    return seq;
-    //}
 
     /** creates a 3D image from timestamped data. */
     osg::Image* createImageSequence( const TileKey& key, ProgressCallback* progress )
     {
-        osg::ImageSequence* seq = new SyncImageSequence(); //osg::ImageSequence();
-
+        osg::ImageSequence* seq = new SyncImageSequence();
+        
         seq->setLoopingMode( osg::ImageStream::LOOPING );
         seq->setLength( _options.secondsPerFrame().value() * (double)_timesVec.size() );
-        seq->play();
+        if ( this->isSequencePlaying() )
+            seq->play();
 
         for( unsigned int r=0; r<_timesVec.size(); ++r )
         {
-            std::string extraAttrs = "TIME=" + _timesVec[r];
+            std::string extraAttrs = std::string("TIME=") + _timesVec[r];
 
-            HTTPResponse response;
+            ReadResult response;
             osgDB::ReaderWriter* reader = fetchTileAndReader( key, extraAttrs, progress, response );
             if ( reader )
             {
-                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( response.getPartStream( 0 ), 0L ); //getOptions() );
+                std::istringstream buf(response.getString());
+                osgDB::ReaderWriter::ReadResult readResult = reader->readImage( buf, _dbOptions.get() );
                 if ( !readResult.error() )
                 {
                     seq->addImage( readResult.getImage() );
@@ -435,13 +466,13 @@ public:
             }
         }
 
+        _sequenceCache.insert( seq );
         return seq;
     }
 
 
     /** override */
-    osg::HeightField* createHeightField( const TileKey& key,
-                                         ProgressCallback* progress)
+    osg::HeightField* createHeightField( const TileKey& key, ProgressCallback* progress)
     {
         osg::Image* image = createImage(key, progress);
         if (!image)
@@ -489,14 +520,75 @@ public:
         return _formatToUse;
     }
 
+
+public: // SequenceControl
+
+    /** Whether the implementation supports these methods */
+    bool supportsSequenceControl() const
+    {
+        return _timesVec.size() > 1;
+    }
+
+    /** Starts playback */
+    void playSequence()
+    {
+        //todo
+        _isPlaying = true;
+    }
+
+    /** Stops playback */
+    void pauseSequence()
+    {
+        //todo
+        _isPlaying = false;
+    }
+
+    /** Seek to a specific frame */
+    void seekToSequenceFrame(unsigned frame)
+    {
+        //todo
+    }
+
+    /** Whether the object is in playback mode */
+    bool isSequencePlaying() const
+    {
+        return _isPlaying;
+    }
+
+    /** Gets data about the current frame in the sequence */
+    const std::vector<SequenceFrameInfo>& getSequenceFrameInfo() const
+    {
+        return _seqFrameInfoVec;
+    }
+
+    /** Index of current frame */
+    int getCurrentSequenceFrameIndex( const osg::FrameStamp* fs ) const
+    {
+        if ( _seqFrameInfoVec.size() == 0 )
+            return 0;
+
+        double len = _options.secondsPerFrame().value() * (double)_timesVec.size();
+        double t   = fmod( fs->getSimulationTime(), len ) / len;
+        return osg::clampBetween(
+            (int)(t * (double)_seqFrameInfoVec.size()), 
+            (int)0, 
+            (int)_seqFrameInfoVec.size()-1);
+    }
+
+
 private:
-    const WMSOptions _options;
-    std::string _formatToUse;
-    std::string _srsToUse;
-    osg::ref_ptr<TileService> _tileService;
-    osg::ref_ptr<const Profile> _profile;
-    std::string _prototype;
-    std::vector<std::string> _timesVec;
+    const WMSOptions                 _options;
+    std::string                      _formatToUse;
+    std::string                      _srsToUse;
+    osg::ref_ptr<TileService>        _tileService;
+    osg::ref_ptr<const Profile>      _profile;
+    std::string                      _prototype;
+    std::vector<std::string>         _timesVec;
+    osg::ref_ptr<osgDB::Options>     _dbOptions;
+    bool                             _isPlaying;
+    std::vector<SequenceFrameInfo>   _seqFrameInfoVec;
+
+    mutable Threading::ThreadSafeObserverSet<osg::ImageSequence> _sequenceCache;
 };
 
 
