@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
-* Copyright 2008-2010 Pelican Mapping
+* Copyright 2008-2013 Pelican Mapping
 * http://osgearth.org
 *
 * osgEarth is free software; you can redistribute it and/or modify
@@ -37,9 +37,10 @@ using namespace geos::operation;
 #define LC "[Geometry] "
 
 
-Geometry::Geometry( const Geometry& rhs, const osg::CopyOp& op ) :
-osg::Vec3dArray( rhs, op )
+Geometry::Geometry( const Geometry& rhs ) :
+osgEarth::MixinVector<osg::Vec3d,osg::Referenced>( rhs )
 {
+    //nop
 }
 
 Geometry::Geometry( int capacity )
@@ -48,7 +49,7 @@ Geometry::Geometry( int capacity )
         reserve( capacity );
 }
 
-Geometry::Geometry( const osg::Vec3dArray* data )
+Geometry::Geometry( const Vec3dVector* data )
 {
     reserve( data->size() );
     insert( begin(), data->begin(), data->end() );
@@ -72,19 +73,24 @@ Geometry::getBounds() const
 Geometry*
 Geometry::cloneAs( const Geometry::Type& newType ) const
 {
-    if ( newType == getType() )
-        return static_cast<Geometry*>( clone() );
+    //if ( newType == getType() )        
+    //    return static_cast<Geometry*>( clone() );
     
     switch( newType )
     {
     case TYPE_POINTSET:
-        return new PointSet( (const osg::Vec3dArray*)this );
+        return new PointSet( &this->asVector() );
     case TYPE_LINESTRING:
-        return new LineString( (const osg::Vec3dArray*)this );
+        return new LineString( &this->asVector() );
     case TYPE_RING:
-        return new Ring( (const osg::Vec3dArray*)this );
+        return new Ring( &this->asVector() );
     case TYPE_POLYGON:
-        return new Polygon( (const osg::Vec3dArray*)this );
+        if ( dynamic_cast<const Polygon*>(this) )
+            return new Polygon( *static_cast<const Polygon*>(this) );
+        else
+            return new Polygon( &this->asVector() );
+    case TYPE_UNKNOWN:
+        return new Geometry( &this->asVector() );
     default:
         break;
     }
@@ -99,8 +105,16 @@ Geometry::toVec3Array() const
     return result;
 }
 
+osg::Vec3dArray*
+Geometry::toVec3dArray() const 
+{
+    osg::Vec3dArray* result = new osg::Vec3dArray( this->size() );
+    std::copy( begin(), end(), result->begin() );
+    return result;
+}
+
 Geometry*
-Geometry::create( Type type, const osg::Vec3dArray* toCopy )
+Geometry::create( Type type, const Vec3dVector* toCopy )
 {
     Geometry* output = 0L;
     switch( type ) {
@@ -154,7 +168,7 @@ Geometry::buffer(double distance,
         //     This seems to only effect the Linux build, Windows works fine
         int geosQuadSegs = params._cornerSegs > 0 
             ? params._cornerSegs
-            : 8;//buffer::BufferParameters::DEFAULT_QUADRANT_SEGMENTS;
+            : 8; //buffer::BufferParameters::DEFAULT_QUADRANT_SEGMENTS;
 
         geom::Geometry* outGeom = NULL;
 
@@ -164,13 +178,21 @@ Geometry::buffer(double distance,
         geosBufferParams.setJoinStyle( geosJoinStyle );
         buffer::BufferBuilder bufBuilder( geosBufferParams );
 
-        if (params._singleSided)
+        try
         {
-            outGeom = bufBuilder.bufferLineSingleSided(inGeom, distance, params._leftSide );
+            if (params._singleSided)
+            {
+                outGeom = bufBuilder.bufferLineSingleSided(inGeom, distance, params._leftSide );
+            }
+            else
+            {
+                outGeom = bufBuilder.buffer(inGeom, distance );
+            }
         }
-        else
+        catch( const util::TopologyException& ex )
         {
-            outGeom = bufBuilder.buffer(inGeom, distance );
+            OE_WARN << LC << "GEOS buffer: " << ex.what() << std::endl;
+            outGeom = 0L;
         }
 
         if ( outGeom )
@@ -178,10 +200,6 @@ Geometry::buffer(double distance,
             output = GEOSUtils::exportGeometry( outGeom );
             outGeom->getFactory()->destroyGeometry( outGeom );
         }
-        else
-        {
-            OE_INFO << LC << "Buffer: no output geometry" << std::endl;
-        }
 
         inGeom->getFactory()->destroyGeometry( inGeom );
     }
@@ -245,24 +263,170 @@ Geometry::crop( const Polygon* cropPoly, osg::ref_ptr<Geometry>& output ) const
 #endif // OSGEARTH_HAVE_GEOS
 }
 
+bool
+Geometry::difference( const Polygon* diffPolygon, osg::ref_ptr<Geometry>& output ) const
+{
+#ifdef OSGEARTH_HAVE_GEOS
+
+    geom::GeometryFactory* f = new geom::GeometryFactory();
+
+    //Create the GEOS Geometries
+    geom::Geometry* inGeom = GEOSUtils::importGeometry( this );
+    geom::Geometry* diffGeom = GEOSUtils::importGeometry( diffPolygon );
+
+    if ( inGeom )
+    {    
+        geom::Geometry* outGeom = 0L;
+        try {
+            outGeom = overlay::OverlayOp::overlayOp(
+                inGeom, diffGeom,
+                overlay::OverlayOp::opDIFFERENCE );
+        }
+        catch( ... ) {
+            outGeom = 0L;
+            OE_NOTICE << LC << "::difference, GEOS overlay op exception, skipping feature" << std::endl;
+        }
+
+        if ( outGeom )
+        {
+            output = GEOSUtils::exportGeometry( outGeom );
+            f->destroyGeometry( outGeom );
+            if ( output.valid() && !output->isValid() )
+            {
+                output = 0L;
+            }
+        }
+    }
+
+    //Destroy the geometry
+    f->destroyGeometry( diffGeom );
+    f->destroyGeometry( inGeom );
+
+    delete f;
+    return output.valid();
+
+#else // OSGEARTH_HAVE_GEOS
+
+    OE_WARN << LC << "Difference failed - GEOS not available" << std::endl;
+    return false;
+
+#endif // OSGEARTH_HAVE_GEOS
+}
+
+osg::Vec3d
+Geometry::localize()
+{
+    osg::Vec3d offset;
+
+    Bounds bounds = getBounds();
+    if ( bounds.isValid() )
+    {      
+        osg::Vec2d center = bounds.center2d();
+        offset.set( center.x(), center.y(), 0 );
+
+        GeometryIterator i( this );
+        while( i.hasMore() )
+        {
+            Geometry* part = i.next();
+            for( Geometry::iterator j = part->begin(); j != part->end(); ++j )
+            {
+                *j = *j - offset;
+            }
+        }
+    }
+
+    return offset;
+}
+
+void
+Geometry::delocalize( const osg::Vec3d& offset )
+{
+    GeometryIterator i( this );
+    while( i.hasMore() )
+    {
+        Geometry* part = i.next();
+        for( Geometry::iterator j = part->begin(); j != part->end(); ++j )
+        {
+            *j = *j + offset;
+        }
+    }
+}
+
+void 
+Geometry::rewind( Orientation orientation )
+{
+    Orientation current = getOrientation();
+    if ( current != orientation && current != ORIENTATION_DEGENERATE && orientation != ORIENTATION_DEGENERATE )
+    {
+        std::reverse( begin(), end() );
+    }
+}
+
+Geometry::Orientation 
+Geometry::getOrientation() const
+{
+    // adjust for a non-open ring:
+    int n = size();
+    while( n > 0 && front() == back() )
+        n--;
+
+    if ( n < 3 )
+        return Geometry::ORIENTATION_DEGENERATE;
+
+    // copy the open vec:
+    std::vector<osg::Vec3d> v;
+    v.reserve( n );
+    std::copy( begin(), begin()+n, std::back_inserter(v) );
+
+    int rmin = 0;
+    double xmin = v[0].x();
+    double ymin = v[0].y();
+    v[0].z() = 0;
+    for( int i=1; i<n; ++i ) {
+        double x = v[i].x();
+        double y = v[i].y();
+        v[i].z() = 0;
+        if ( y > ymin )
+            continue;
+        if ( y == ymin ) {
+            if (x  < xmin )
+                continue;
+        }
+        rmin = i;
+        xmin = x;
+        ymin = y;
+    }
+
+    int rmin_less_1 = rmin-1 >= 0 ? rmin-1 : n-1;
+    int rmin_plus_1 = rmin+1 < n ? rmin+1 : 0;
+
+    osg::Vec3 in = v[rmin] - v[rmin_less_1]; in.normalize();
+    osg::Vec3 out = v[rmin_plus_1] - v[rmin]; out.normalize();
+    osg::Vec3 cross = in ^ out;
+
+    return
+        cross.z() < 0.0 ? Geometry::ORIENTATION_CW :
+        cross.z() > 0.0 ? Geometry::ORIENTATION_CCW :
+        Geometry::ORIENTATION_DEGENERATE;
+}
 
 //----------------------------------------------------------------------------
 
-PointSet::PointSet( const PointSet& rhs, const osg::CopyOp& op ) :
-Geometry( rhs, op )
+PointSet::PointSet( const PointSet& rhs ) :
+Geometry( rhs )
 {
     //nop
 }
 
 //----------------------------------------------------------------------------
 
-LineString::LineString( const LineString& rhs, const osg::CopyOp& op ) :
-Geometry( rhs, op )
+LineString::LineString( const LineString& rhs ) :
+Geometry( rhs )
 {
     //nop
 }
 
-LineString::LineString( const osg::Vec3dArray* data ) :
+LineString::LineString( const Vec3dVector* data ) :
 Geometry( data )
 {
     //nop
@@ -302,13 +466,13 @@ LineString::getSegment(double length, osg::Vec3d& start, osg::Vec3d& end)
 
 //----------------------------------------------------------------------------
 
-Ring::Ring( const Ring& rhs, const osg::CopyOp& op ) :
-Geometry( rhs, op )
+Ring::Ring( const Ring& rhs ) :
+Geometry( rhs )
 {
     //nop
 }
 
-Ring::Ring( const osg::Vec3dArray* data ) :
+Ring::Ring( const Vec3dVector* data ) :
 Geometry( data )
 {
     open();
@@ -319,7 +483,7 @@ Ring::cloneAs( const Geometry::Type& newType ) const
 {
     if ( newType == TYPE_LINESTRING )
     {
-        LineString* line = new LineString( (osg::Vec3dArray*)this );
+        LineString* line = new LineString( &this->asVector() );
         if ( line->size() > 1 && line->front() != line->back() )
             line->push_back( front() );
         return line;
@@ -327,54 +491,6 @@ Ring::cloneAs( const Geometry::Type& newType ) const
     else return Geometry::cloneAs( newType );
 }
 
-Ring::Orientation 
-Ring::getOrientation() const
-{
-    // adjust for a non-open ring:
-    int n = size();
-    while( n > 0 && front() == back() )
-        n--;
-
-    if ( n < 3 )
-        return Ring::ORIENTATION_DEGENERATE;
-
-    // copy the open vec:
-    std::vector<osg::Vec3d> v;
-    v.reserve( n );
-    std::copy( begin(), begin()+n, std::back_inserter(v) );
-
-    int rmin = 0;
-    double xmin = v[0].x();
-    double ymin = v[0].y();
-    v[0].z() = 0;
-    for( int i=1; i<n; ++i ) {
-        double x = v[i].x();
-        double y = v[i].y();
-        v[i].z() = 0;
-        if ( y > ymin )
-            continue;
-        if ( y == ymin ) {
-            if (x  < xmin )
-                continue;
-        }
-        rmin = i;
-        xmin = x;
-        ymin = y;
-    }
-
-    int rmin_less_1 = rmin-1 >= 0 ? rmin-1 : n-1;
-    int rmin_plus_1 = rmin+1 < n ? rmin+1 : 0;
-
-    osg::Vec3 in = v[rmin] - v[rmin_less_1]; in.normalize();
-    osg::Vec3 out = v[rmin_plus_1] - v[rmin]; out.normalize();
-    osg::Vec3 cross = in ^ out;
-
-    return
-        cross.z() < 0.0 ? Ring::ORIENTATION_CW :
-        cross.z() > 0.0 ? Ring::ORIENTATION_CCW :
-        Ring::ORIENTATION_DEGENERATE;
-}
-
 // ensures that the first and last points are not idential.
 void 
 Ring::open()
@@ -383,29 +499,59 @@ Ring::open()
         erase( end()-1 );
 }
 
+// gets the signed area.
+double
+Ring::getSignedArea2D() const
+{
+    const_cast<Ring*>(this)->open();
+
+    double sum = 0.0;
+
+    for( unsigned i=0; i<size(); ++i )
+    {
+        const osg::Vec3d& p0 = (*this)[0];
+        const osg::Vec3d& p1 = i+1 < size() ? (*this)[i+1] : (*this)[0];
+        sum += p0.x()*p1.y() - p1.x()*p0.y();
+    }
+    return .5*sum;
+}
+
 // opens and rewinds the polygon to the specified orientation.
 void 
 Ring::rewind( Orientation orientation )
 {
     open();
-    Orientation current = getOrientation();
-    if ( current != orientation && current != ORIENTATION_DEGENERATE && orientation != ORIENTATION_DEGENERATE )
+    Geometry::rewind( orientation );
+}
+
+// point-in-polygon test
+bool
+Ring::contains2D( double x, double y ) const
+{
+    bool result = false;
+    const Ring& poly = *this;
+    for( unsigned i=0, j=size()-1; i<size(); j = i++ )
     {
-        std::reverse( begin(), end() );
+        if ((((poly[i].y() <= y) && (y < poly[j].y())) ||
+            ((poly[j].y() <= y) && (y < poly[i].y()))) &&
+            (x < (poly[j].x()-poly[i].x()) * (y-poly[i].y())/(poly[j].y()-poly[i].y())+poly[i].x()))
+        {
+            result = !result;
+        }
     }
+    return result;
 }
 
 //----------------------------------------------------------------------------
 
-Polygon::Polygon( const Polygon& rhs, const osg::CopyOp& op ) :
-Ring( rhs, op )
+Polygon::Polygon( const Polygon& rhs ) :
+Ring( rhs )
 {
-    //_boundary = rhs._boundary.valid() ? osg::clone<Ring>( rhs._boundary.get(), op ) : 0L;
     for( RingCollection::const_iterator r = rhs._holes.begin(); r != rhs._holes.end(); ++r )
-        _holes.push_back( osg::clone<Ring>( r->get(), op ) );
+        _holes.push_back( new Ring(*r->get()) );
 }
 
-Polygon::Polygon( const osg::Vec3dArray* data ) :
+Polygon::Polygon( const Vec3dVector* data ) :
 Ring( data )
 {
     //nop
@@ -414,19 +560,45 @@ Ring( data )
 int
 Polygon::getTotalPointCount() const
 {
-    int total = Ring::getTotalPointCount(); //_boundary.valid() ? _boundary->size() : 0;
+    int total = Ring::getTotalPointCount();
     for( RingCollection::const_iterator i = _holes.begin(); i != _holes.end(); ++i )
         total += i->get()->getTotalPointCount();
     return total;
 }
 
+bool
+Polygon::contains2D( double x, double y ) const
+{
+    // first check the outer ring
+    if ( !Ring::contains2D(x, y) )
+        return false;
+
+    // then check each inner ring (holes). Point has to be inside the outer ring, 
+    // but NOT inside any of the holes
+    for( RingCollection::const_iterator i = _holes.begin(); i != _holes.end(); ++i )
+    {
+        if ( i->get()->contains2D(x, y) )
+            return false;
+    }
+
+    return true;
+}
+
+void
+Polygon::open() 
+{
+    Ring::open();
+    for( RingCollection::const_iterator i = _holes.begin(); i != _holes.end(); ++i )
+        (*i)->open();
+}
+
 //----------------------------------------------------------------------------
 
-MultiGeometry::MultiGeometry( const MultiGeometry& rhs, const osg::CopyOp& op ) :
-Geometry( rhs, op )
+MultiGeometry::MultiGeometry( const MultiGeometry& rhs ) :
+Geometry( rhs )
 {
     for( GeometryCollection::const_iterator i = rhs._parts.begin(); i != rhs._parts.end(); ++i )
-        _parts.push_back( osg::clone<Geometry>( i->get() ) );
+        _parts.push_back( i->get()->clone() ); //i->clone() ); //osg::clone<Geometry>( i->get() ) );
 }
 
 MultiGeometry::MultiGeometry( const GeometryCollection& parts ) :
@@ -451,6 +623,15 @@ MultiGeometry::getTotalPointCount() const
     return total;
 }
 
+unsigned
+MultiGeometry::getNumGeometries() const
+{
+    unsigned total = 0;
+    for( GeometryCollection::const_iterator i = _parts.begin(); i != _parts.end(); ++i )
+        total += i->get()->getNumGeometries();
+    return total;
+}
+
 Bounds
 MultiGeometry::getBounds() const
 {
@@ -468,7 +649,7 @@ MultiGeometry::cloneAs( const Geometry::Type& newType ) const
     MultiGeometry* multi = new MultiGeometry();
     for( GeometryCollection::const_iterator i = _parts.begin(); i != _parts.end(); ++i )
     {
-        Geometry* part = i->get()->cloneAs( newType );
+        Geometry* part = i->get()->cloneAs( i->get()->getType() );
         if ( part ) multi->getComponents().push_back( part );
     }
     return multi;
@@ -489,12 +670,22 @@ MultiGeometry::isValid() const
     return valid;
 }
 
+// opens and rewinds the polygon to the specified orientation.
+void 
+MultiGeometry::rewind( Orientation orientation )
+{
+    for( GeometryCollection::const_iterator i = _parts.begin(); i != _parts.end(); ++i )
+    {
+        i->get()->rewind( orientation );
+    }
+}
+
 //----------------------------------------------------------------------------
 
-GeometryIterator::GeometryIterator( Geometry* geom ) :
+GeometryIterator::GeometryIterator( Geometry* geom, bool holes ) :
 _next( 0L ),
 _traverseMulti( true ),
-_traversePolyHoles( true )
+_traversePolyHoles( holes )
 {
     if ( geom )
     {
@@ -549,10 +740,10 @@ GeometryIterator::fetchNext()
 
 //----------------------------------------------------------------------------
 
-ConstGeometryIterator::ConstGeometryIterator( const Geometry* geom ) :
+ConstGeometryIterator::ConstGeometryIterator( const Geometry* geom, bool holes ) :
 _next( 0L ),
 _traverseMulti( true ),
-_traversePolyHoles( true )
+_traversePolyHoles( holes )
 {
     if ( geom )
     {
@@ -605,3 +796,34 @@ ConstGeometryIterator::fetchNext()
     }    
 }
 
+//----------------------------------------------------------------------------
+
+ConstSegmentIterator::ConstSegmentIterator( const Geometry* verts, bool forceClosedLoop ) :
+_verts(&verts->asVector()),
+_closeLoop(forceClosedLoop),
+_iter(verts->begin())
+{
+    _done = _verts->size() < 2;
+
+    if ( !forceClosedLoop )
+    {
+        _closeLoop = dynamic_cast<const Ring*>(verts) != 0L;
+    }
+}
+
+Segment
+ConstSegmentIterator::next()
+{
+    osg::Vec3d p0 = *_iter++;
+    if ( _iter == _verts->end() ) 
+    {
+        _iter = _verts->begin();
+        _done = true;
+    }
+    else if ( _iter+1 == _verts->end() && !_closeLoop )
+    {
+        _done = true;
+    }
+
+    return Segment( p0, *_iter );
+}
