@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -17,9 +17,10 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>
  */
 #include <osgEarthSymbology/MeshSubdivider>
-#include <osgEarthSymbology/LineFunctor>
+#include <osgEarth/LineFunctor>
 #include <osgEarth/GeoMath>
 #include <osg/TriangleFunctor>
+#include <osg/TriangleIndexFunctor>
 //#include <osgUtil/MeshOptimizers>
 #include <climits>
 #include <queue>
@@ -34,6 +35,15 @@ using namespace osgEarth::Symbology;
 
 namespace
 {
+    // returns the geocentric bisection vector
+    osg::Vec3d
+    bisector( const osg::Vec3d& v0, const osg::Vec3d& v1 )
+    {
+        osg::Vec3d f = (v0+v1)*0.5;
+        f.normalize();
+        return f * 0.5*(v0.length() + v1.length());
+    }
+
     // convert geocenric coords to spherical geodetic coords in radians.
     void
     geocentricToGeodetic( const osg::Vec3d& g, osg::Vec2d& out_geod )
@@ -53,27 +63,36 @@ namespace
             out_mid.set( 0.5*((g0.x()+2*osg::PI)+g1.x()), 0.5*(g0.y()+g1.y()) );
         else
            out_mid.set( 0.5*(g0.x()+(g1.x()+2*osg::PI)), 0.5*(g0.y()+g1.y()) );
+
+        //GeoMath::midpoint(g0.y(), g0.x(), g1.y(), g1.x(), out_mid.y(), out_mid.x());
     }
 
     // finds the midpoint between two geocentric coordinates. We have to convert
     // back to geographic in order to get the correct interpolation. Spherical
     // conversion is good enough
     osg::Vec3d
-    geocentricMidpoint( const osg::Vec3d& v0, const osg::Vec3d& v1 )
+    geocentricMidpoint( const osg::Vec3d& v0, const osg::Vec3d& v1, GeoInterpolation interp )
     {
-        // geocentric to spherical:
-        osg::Vec2d g0, g1;
-        geocentricToGeodetic(v0, g0);
-        geocentricToGeodetic(v1, g1);
+        if ( interp == GEOINTERP_GREAT_CIRCLE )
+        {
+            return bisector(v0, v1);
+        }
+        else
+        {
+            // geocentric to spherical:
+            osg::Vec2d g0, g1;
+            geocentricToGeodetic(v0, g0);
+            geocentricToGeodetic(v1, g1);
 
-        osg::Vec2d mid;
-        geodeticMidpoint(g0, g1, mid);
+            osg::Vec2d mid;
+            geodeticMidpoint(g0, g1, mid);
 
-        double size = 0.5*(v0.length() + v1.length());
+            double size = 0.5*(v0.length() + v1.length());
 
-        // spherical to geocentric:
-        double sin_lat = sin(mid.y());
-        return osg::Vec3d( cos(mid.x())*sin_lat, sin(mid.x())*sin_lat, cos(mid.y()) ) * size;
+            // spherical to geocentric:
+            double sin_lat = sin(mid.y());
+            return osg::Vec3d( cos(mid.x())*sin_lat, sin(mid.x())*sin_lat, cos(mid.y()) ) * size;
+        }
     }
 
     // the approximate surface-distance between two geocentric points (spherical)
@@ -84,16 +103,7 @@ namespace
         geocentricToGeodetic(v0, g0);
         geocentricToGeodetic(v1, g1);
         return GeoMath::distance( v0.y(), v0.x(), v1.y(), v1.x() );
-    }    
-
-    // returns the geocentric bisection vector
-    osg::Vec3d
-    bisector( const osg::Vec3d& v0, const osg::Vec3d& v1 )
-    {
-        osg::Vec3d f = (v0+v1)*0.5;
-        f.normalize();
-        return f * 0.5*(v0.length() + v1.length());
-    }    
+    }
 
     // the angle between two 3d vectors
     double
@@ -109,33 +119,91 @@ namespace
     struct Triangle
     {
         Triangle() { }
-        Triangle(GLuint i0, GLuint i1, GLuint i2) : _i0(i0), _i1(i1), _i2(i2) { }
+        Triangle(GLuint i0, GLuint i1, GLuint i2) : 
+        _i0(i0), _i1(i1), _i2(i2){}                 
         GLuint _i0, _i1, _i2;
     };
 
     typedef std::queue<Triangle> TriangleQueue;
     typedef std::vector<Triangle> TriangleVector;
+    
 
     struct TriangleData
     {
-        typedef std::map<osg::Vec3,GLuint> VertMap;
+        typedef std::map<osg::Vec3,GLuint> VertMap;        
         VertMap _vertMap;
-        osg::Vec3Array* _verts;
+        osg::Vec3Array* _sourceVerts;
+        osg::Vec4Array* _sourceColors;
+        osg::Vec2Array* _sourceTexCoords;
+        osg::Vec3Array* _sourceNormals;
+        osg::ref_ptr<osg::Vec3Array> _verts;        
+        osg::ref_ptr<osg::Vec4Array> _colors;
+        osg::ref_ptr<osg::Vec2Array> _texcoords;
+        osg::ref_ptr<osg::Vec3Array> _normals;
         TriangleQueue _tris;
         
         TriangleData()
+        {            
+            _verts           = new osg::Vec3Array();             
+            _sourceVerts     = 0;
+            _sourceColors    = 0;
+            _sourceTexCoords = 0;
+            _sourceNormals   = 0;
+        }       
+
+        void setSourceVerts(osg::Vec3Array* sourceVerts )
         {
-            _verts = new osg::Vec3Array();
+            _sourceVerts = sourceVerts;
         }
 
-        GLuint record( const osg::Vec3& v )
+        void setSourceColors(osg::Vec4Array* sourceColors)
+        {
+            if ( sourceColors )
+            {
+                _sourceColors = sourceColors;
+                _colors       = new osg::Vec4Array();
+            }
+        }
+
+        void setSourceTexCoords(osg::Vec2Array* sourceTexCoords)
+        {
+            if ( sourceTexCoords )
+            {
+                _sourceTexCoords = sourceTexCoords;
+                _texcoords       = new osg::Vec2Array();
+            }
+        }
+
+        void setSourceNormals(osg::Vec3Array* sourceNormals)
+        {
+            if ( sourceNormals )
+            {
+                _sourceNormals = sourceNormals;
+                _normals       = new osg::Vec3Array();
+            }
+        }
+
+        GLuint record( const osg::Vec3& v, const osg::Vec2f& t, const osg::Vec4f& c, const osg::Vec3& n )
         {
             VertMap::iterator i = _vertMap.find(v);
             if ( i == _vertMap.end() )
             {
                 GLuint index = _verts->size();
-                _verts->push_back(v);
+                _verts->push_back(v);                
                 _vertMap[v] = index;
+                //Only push back the texture coordinate if it's valid
+                if (_texcoords)
+                {
+                    _texcoords->push_back( t );
+                }
+                if (_colors)
+                {
+                    _colors->push_back( c );
+                }
+                if ( _normals )
+                {
+                    _normals->push_back( n );
+                }
                 return index;
             }
             else
@@ -143,10 +211,39 @@ namespace
                 return i->second;
             }
         }
-        
-        void operator()( const osg::Vec3& v0, const osg::Vec3& v1, const osg::Vec3& v2, bool temp )
+       
+
+        void operator() (unsigned int p1, unsigned int p2, unsigned int p3)
         {
-            _tris.push( Triangle(record(v0), record(v1), record(v2)) );
+            const osg::Vec3 v0 = (*_sourceVerts)[p1];
+            const osg::Vec3 v1 = (*_sourceVerts)[p2];
+            const osg::Vec3 v2 = (*_sourceVerts)[p3];   
+
+            osg::Vec2 t0, t1, t2;
+            if (_sourceTexCoords)
+            {
+                t0 = (*_sourceTexCoords)[p1];
+                t1 = (*_sourceTexCoords)[p2];
+                t2 = (*_sourceTexCoords)[p3];
+            }
+            
+            osg::Vec4 c0, c1, c2;
+            if (_sourceColors)
+            {
+                c0 = (*_sourceColors)[p1];
+                c1 = (*_sourceColors)[p2];
+                c2 = (*_sourceColors)[p3];
+            }
+
+            osg::Vec3 n0, n1, n2;
+            if (_sourceNormals)
+            {
+                n0 = (*_sourceNormals)[p1];
+                n1 = (*_sourceNormals)[p2];
+                n2 = (*_sourceNormals)[p3];
+            }
+
+            _tris.push( Triangle(record(v0, t0, c0, n0), record(v1, t1, c1, n1), record(v2, t2, c2, n2)) );
         }
     };      
 
@@ -219,29 +316,70 @@ namespace
         GLuint _i0, _i1;
     };
 
-    typedef std::queue<Line> LineQueue;
+    typedef std::queue<Line>  LineQueue;
     typedef std::vector<Line> LineVector;
 
     struct LineData
     {
-        typedef std::map<osg::Vec3,GLuint> VertMap;
+        typedef std::map<osg::Vec3,GLuint> VertMap;        
         VertMap _vertMap;
-        osg::Vec3Array* _verts;
+        osg::Vec3Array* _sourceVerts;
+        osg::Vec4Array* _sourceColors;
+        osg::Vec2Array* _sourceTexCoords;
+        osg::ref_ptr<osg::Vec3Array> _verts;        
+        osg::ref_ptr<osg::Vec4Array> _colors;
+        osg::ref_ptr<osg::Vec2Array> _texcoords;
         LineQueue _lines;
         
         LineData()
+        {            
+            _verts           = new osg::Vec3Array();      
+            _colors          = 0;
+            _texcoords       = 0;
+            _sourceVerts     = 0;
+            _sourceColors    = 0;
+            _sourceTexCoords = 0;
+        }       
+
+        void setSourceVerts(osg::Vec3Array* sourceVerts )
         {
-            _verts = new osg::Vec3Array();
+            _sourceVerts = sourceVerts;
         }
 
-        GLuint record( const osg::Vec3& v )
+        void setSourceColors(osg::Vec4Array* sourceColors )
+        {
+            if ( sourceColors )
+            {
+                _sourceColors = sourceColors;
+                _colors       = new osg::Vec4Array();
+            }
+        }
+
+        void setSourceTexCoords(osg::Vec2Array* sourceTexCoords)
+        {
+            if ( sourceTexCoords )
+            {
+                _sourceTexCoords = sourceTexCoords;
+                _texcoords       = new osg::Vec2Array();
+            }
+        }
+
+        GLuint record( const osg::Vec3& v, const osg::Vec2f& t, const osg::Vec4f& c )
         {
             VertMap::iterator i = _vertMap.find(v);
             if ( i == _vertMap.end() )
             {
                 GLuint index = _verts->size();
-                _verts->push_back(v);
+                _verts->push_back(v);                
                 _vertMap[v] = index;
+                if (_texcoords)
+                {
+                    _texcoords->push_back( t );
+                }
+                if (_colors)
+                {
+                    _colors->push_back( c );
+                }
                 return index;
             }
             else
@@ -249,12 +387,30 @@ namespace
                 return i->second;
             }
         }
-        
-        void operator()( const osg::Vec3& v0, const osg::Vec3& v1, bool temp )
+       
+
+        void line(unsigned p0, unsigned p1)
         {
-            _lines.push( Line( record(v0), record(v1) ) );
+            const osg::Vec3 v0 = (*_sourceVerts)[p0];
+            const osg::Vec3 v1 = (*_sourceVerts)[p1];   
+
+            osg::Vec2 t0, t1;
+            if (_sourceTexCoords)
+            {
+                t0 = (*_sourceTexCoords)[p0];
+                t1 = (*_sourceTexCoords)[p1];
+            }
+            
+            osg::Vec4 c0, c1;
+            if (_sourceColors)
+            {
+                c0 = (*_sourceColors)[p0];
+                c1 = (*_sourceColors)[p1];
+            }
+
+            _lines.push( Line(record(v0,t0,c0), record(v1,t1,c1)) );
         }
-    };       
+    };         
     
     /**
      * Populates the geometry object with a collection of index elements primitives.
@@ -306,14 +462,19 @@ namespace
      * the data in the Geometry object with the new vertex and primitive data.
      */
     void subdivideLines(
-        double granularity,
-        osg::Geometry& geom,
-        const osg::Matrixd& W2L, // world=>local xform
-        const osg::Matrixd& L2W, // local=>world xform
-        unsigned int maxElementsPerEBO )
+        double               granularity,
+        GeoInterpolation     interp,
+        osg::Geometry&       geom,
+        const osg::Matrixd&  W2L, // world=>local xform
+        const osg::Matrixd&  L2W, // local=>world xform
+        unsigned int         maxElementsPerEBO )
     {
         // collect all the line segments in the geometry.
-        LineFunctor<LineData> data;
+        LineIndexFunctor<LineData> data;
+        data.setSourceVerts( static_cast<osg::Vec3Array*>(geom.getVertexArray()) );
+        if ( geom.getColorBinding() == osg::Geometry::BIND_PER_VERTEX )
+            data.setSourceColors( static_cast<osg::Vec4Array*>(geom.getColorArray()) );
+        //LineFunctor<LineData> data;
         geom.accept( data );
     
         int numLinesIn = data._lines.size();
@@ -334,7 +495,22 @@ namespace
 
             if ( g0 > granularity )
             {
-                data._verts->push_back( geocentricMidpoint(v0_w, v1_w) * W2L );
+                data._verts->push_back( geocentricMidpoint(v0_w, v1_w, interp) * W2L );
+
+                if ( data._colors )
+                {
+                    const osg::Vec4f& c0 = (*data._colors)[line._i0];
+                    const osg::Vec4f& c1 = (*data._colors)[line._i1];
+                    data._colors->push_back( (c0 + c1) / 2.0 );
+                }
+
+                if ( data._texcoords )
+                {
+                    const osg::Vec2& t0 = (*data._texcoords)[line._i0];
+                    const osg::Vec2& t1 = (*data._texcoords)[line._i1];
+                    data._texcoords->push_back( (t0 + t1) / 2.0 );
+                }
+
                 GLuint i = data._verts->size()-1;
 
                 data._lines.push( Line( line._i0, i ) );
@@ -354,6 +530,13 @@ namespace
 
             // set the new VBO.
             geom.setVertexArray( data._verts );
+            if ( geom.getVertexArray()->getVertexBufferObject() && data._verts->getVertexBufferObject() )
+            {
+                data._verts->getVertexBufferObject()->setUsage( geom.getVertexArray()->getVertexBufferObject()->getUsage() );
+            }
+
+            if ( data._colors )
+                geom.setColorArray( data._colors );
 
             if ( data._verts->size() < 256 )
                 populateLines<osg::DrawElementsUByte,GLubyte>( geom, done, maxElementsPerEBO );
@@ -377,17 +560,27 @@ namespace
      * (c) Copyright 2010 Patrick Cozzi and Deron Ohlarik, MIT License.
      */
     void subdivideTriangles(
-        double granularity,
-        osg::Geometry& geom,
-        const osg::Matrixd& W2L, // world=>local xform
-        const osg::Matrixd& L2W, // local=>world xform
-        unsigned int maxElementsPerEBO )
+        double               granularity,
+        GeoInterpolation     interp,
+        osg::Geometry&       geom,
+        const osg::Matrixd&  W2L, // world=>local xform
+        const osg::Matrixd&  L2W, // local=>world xform
+        unsigned int         maxElementsPerEBO )
     {
+        
         // collect all the triangled in the geometry.
-        osg::TriangleFunctor<TriangleData> data;
-        geom.accept( data );
+        osg::TriangleIndexFunctor<TriangleData> data;;
+        data.setSourceVerts(dynamic_cast<osg::Vec3Array*>(geom.getVertexArray()));
+        data.setSourceTexCoords(dynamic_cast<osg::Vec2Array*>(geom.getTexCoordArray(0)));
+        if ( geom.getColorBinding() == osg::Geometry::BIND_PER_VERTEX )
+            data.setSourceColors(dynamic_cast<osg::Vec4Array*>(geom.getColorArray()));
+        if ( geom.getNormalBinding() == osg::Geometry::BIND_PER_VERTEX )
+            data.setSourceNormals(dynamic_cast<osg::Vec3Array*>(geom.getNormalArray()));
 
-        int numTrisIn = data._tris.size();
+        //TODO normals
+        geom.accept( data );
+        
+        int numTrisIn = data._tris.size();        
 
         TriangleVector done;
         done.reserve(2.0 * data._tris.size());
@@ -405,6 +598,30 @@ namespace
             osg::Vec3d v1_w = (*data._verts)[tri._i1] * L2W;
             osg::Vec3d v2_w = (*data._verts)[tri._i2] * L2W;
 
+            osg::Vec2 t0,t1,t2;
+            if ( data._texcoords.valid() )
+            {
+                t0 = (*data._texcoords)[tri._i0];
+                t1 = (*data._texcoords)[tri._i1];
+                t2 = (*data._texcoords)[tri._i2];
+            }
+
+            osg::Vec4f c0,c1,c2;
+            if ( data._colors.valid() )
+            {
+                c0 = (*data._colors)[tri._i0];
+                c1 = (*data._colors)[tri._i1];
+                c2 = (*data._colors)[tri._i2];
+            }
+
+            osg::Vec3 n0,n1,n2;
+            if ( data._normals.valid() )
+            {
+                n0 = (*data._normals)[tri._i0];
+                n1 = (*data._normals)[tri._i1];
+                n2 = (*data._normals)[tri._i2];
+            }
+
             double g0 = angleBetween(v0_w, v1_w);
             double g1 = angleBetween(v1_w, v2_w);
             double g2 = angleBetween(v2_w, v0_w);
@@ -420,7 +637,13 @@ namespace
                     GLuint i;
                     if ( ei == edges.end() )
                     {
-                        data._verts->push_back( geocentricMidpoint(v0_w, v1_w) * W2L );
+                        data._verts->push_back( geocentricMidpoint(v0_w, v1_w, interp) * W2L );
+                        if ( data._colors.valid() )
+                            data._colors->push_back( (c0 + c1) / 2.0f );
+                        if ( data._texcoords.valid() )
+                            data._texcoords->push_back( (t0 + t1) / 2.0f );
+                        if ( data._normals.valid() )
+                            data._normals->push_back( (n0 + n1) / 2.0f );
                         i = data._verts->size() - 1;
                         edges[edge] = i;
                     }
@@ -440,7 +663,13 @@ namespace
                     GLuint i;
                     if ( ei == edges.end() )
                     {
-                        data._verts->push_back( geocentricMidpoint(v1_w, v2_w) * W2L );
+                        data._verts->push_back( geocentricMidpoint(v1_w, v2_w, interp) * W2L );
+                        if ( data._colors.valid() )
+                            data._colors->push_back( (c1 + c2) / 2.0f );
+                        if ( data._texcoords.valid() )
+                            data._texcoords->push_back( (t1 + t2) / 2.0f );
+                        if ( data._normals.valid() )
+                            data._normals->push_back( (n1 + n2) / 2.0f );
                         i = data._verts->size() - 1;
                         edges[edge] = i;
                     }
@@ -460,7 +689,13 @@ namespace
                     GLuint i;
                     if ( ei == edges.end() )
                     {
-                        data._verts->push_back( geocentricMidpoint(v2_w, v0_w) * W2L );
+                        data._verts->push_back( geocentricMidpoint(v2_w, v0_w, interp) * W2L );
+                        if ( data._colors.valid() )
+                            data._colors->push_back( (c2 + c0) / 2.0f );
+                        if ( data._texcoords.valid() )
+                            data._texcoords->push_back( (t2 + t0) / 2.0f );
+                        if ( data._normals.valid() )
+                            data._normals->push_back( (n2 + n0) / 2.0f );
                         i = data._verts->size() - 1;
                         edges[edge] = i;
                     }
@@ -487,7 +722,18 @@ namespace
                 geom.removePrimitiveSet( 0 );
 
             // set the new VBO.
-            geom.setVertexArray( data._verts );
+            geom.setVertexArray( data._verts.get() );
+            if ( geom.getVertexArray()->getVertexBufferObject() && data._verts->getVertexBufferObject() )
+            {
+                data._verts->getVertexBufferObject()->setUsage( geom.getVertexArray()->getVertexBufferObject()->getUsage() );
+            }
+
+            if ( data._colors.valid() )
+                geom.setColorArray( data._colors.get() );
+            if ( data._texcoords.valid() )
+                geom.setTexCoordArray(0, data._texcoords.get() );
+            if ( data._normals.valid() )
+                geom.setNormalArray( data._normals.get() );
 
             if ( data._verts->size() < 256 )
                 populateTriangles<osg::DrawElementsUByte,GLubyte>( geom, done, maxElementsPerEBO );
@@ -499,11 +745,12 @@ namespace
     }
 
     void subdivide(
-        double granularity,
-        osg::Geometry& geom,
-        const osg::Matrixd& W2L, // world=>local xform
-        const osg::Matrixd& L2W, // local=>world xform
-        unsigned int maxElementsPerEBO )
+        double               granularity,
+        GeoInterpolation     interp,
+        osg::Geometry&       geom,
+        const osg::Matrixd&  W2L, // world=>local xform
+        const osg::Matrixd&  L2W, // local=>world xform
+        unsigned int         maxElementsPerEBO )
     {
         GLenum mode = geom.getPrimitiveSet(0)->getMode();
 
@@ -512,11 +759,11 @@ namespace
 
         if ( mode == GL_LINES || mode == GL_LINE_STRIP || mode == GL_LINE_LOOP )
         {
-            subdivideLines( granularity, geom, W2L, L2W, maxElementsPerEBO );
+            subdivideLines( granularity, interp, geom, W2L, L2W, maxElementsPerEBO );
         }
         else
         {
-            subdivideTriangles( granularity, geom, W2L, L2W, maxElementsPerEBO );
+            subdivideTriangles( granularity, interp, geom, W2L, L2W, maxElementsPerEBO );
 
             //osgUtil::VertexCacheVisitor cacheOptimizer;
             //cacheOptimizer.optimizeVertices( geom );
@@ -542,10 +789,10 @@ _maxElementsPerEBO( INT_MAX )
 }
 
 void
-MeshSubdivider::run(double granularity, osg::Geometry& geom)
+MeshSubdivider::run(osg::Geometry& geom, double granularity, GeoInterpolation interp)
 {
     if ( geom.getNumPrimitiveSets() < 1 )
         return;
 
-    subdivide( granularity, geom, _world2local, _local2world, _maxElementsPerEBO );
+    subdivide( granularity, interp, geom, _world2local, _local2world, _maxElementsPerEBO );
 }
