@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2008-2010 Pelican Mapping
+ * Copyright 2008-2013 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -18,9 +18,23 @@
  */
 #include <osgEarthSymbology/Symbol>
 
+using namespace osgEarth;
 using namespace osgEarth::Symbology;
 
-Symbol::Symbol()
+//------------------------------------------------------------------------
+
+Symbol::Symbol( const Config& conf )
 {
-    //NOP
+    _uriContext = URIContext(conf.referrer());
+}
+
+bool
+Symbol::match(const std::string& s, const char* reservedWord)
+{
+    if ( s.compare(reservedWord) == 0 ) return true;
+    //if ( s == reservedWord ) return true;
+    std::string temp1 = toLower(s), temp2 = toLower(reservedWord);
+    replaceIn(temp1, "_", "-");
+    replaceIn(temp2, "_", "-");
+    return temp1.compare(temp2) == 0;
 }
